import dataclasses as dc
import typing as ta

from ... import check
from ... import collections as col
from ... import lang
from ... import reflect as rfl
from .configs import EMPTY_CONFIG_REGISTRY
from .configs import Configs
from .errors import UnhandledTypeError
from .options import Option
from .overrides import ReflectOverride


if ta.TYPE_CHECKING:
    from .types import Marshaler
    from .types import MarshalerFactory
    from .types import Unmarshaler
    from .types import UnmarshalerFactory
    from .values import Value


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True, kw_only=True)
class BaseContext(lang.Abstract, lang.Sealed):
    configs: Configs = EMPTY_CONFIG_REGISTRY
    options: col.TypeMap[Option] = col.TypeMap()

    def _reflect(self, o: ta.Any) -> rfl.Type:
        def override(o):
            if (ovr := self.configs.get_of(o, ReflectOverride)):
                return ovr[-1].rty
            return None

        return rfl.Reflector(override=override).type(o)


#


@dc.dataclass(frozen=True, kw_only=True)
class MarshalFactoryContext(BaseContext, lang.Final):
    marshaler_factory: ta.Optional['MarshalerFactory'] = None

    def make_marshaler(self, o: ta.Any) -> 'Marshaler':
        rty = self._reflect(o)
        fac = check.not_none(self.marshaler_factory)
        if (m := fac.make_marshaler(self, rty)) is None:
            raise UnhandledTypeError(rty)  # noqa
        return m()


@dc.dataclass(frozen=True, kw_only=True)
class UnmarshalFactoryContext(BaseContext, lang.Final):
    unmarshaler_factory: ta.Optional['UnmarshalerFactory'] = None

    def make_unmarshaler(self, o: ta.Any) -> 'Unmarshaler':
        rty = self._reflect(o)
        fac = check.not_none(self.unmarshaler_factory)
        if (m := fac.make_unmarshaler(self, rty)) is None:
            raise UnhandledTypeError(rty)  # noqa
        return m()


#


@dc.dataclass(frozen=True, kw_only=True)
class MarshalContext(BaseContext, lang.Final):
    marshal_factory_context: MarshalFactoryContext

    def marshal(self, obj: ta.Any, ty: ta.Any | None = None) -> 'Value':
        return self.marshal_factory_context.make_marshaler(ty if ty is not None else type(obj)).marshal(self, obj)


@dc.dataclass(frozen=True, kw_only=True)
class UnmarshalContext(BaseContext, lang.Final):
    unmarshal_factory_context: UnmarshalFactoryContext

    @ta.overload
    def unmarshal(self, v: 'Value', ty: type[T]) -> T:
        ...

    @ta.overload
    def unmarshal(self, v: 'Value', ty: ta.Any) -> ta.Any:
        ...

    def unmarshal(self, v, ty):
        return self.unmarshal_factory_context.make_unmarshaler(ty).unmarshal(self, v)



from .. import lang as _lang  # noqa


with _lang.auto_proxy_init(globals()):
    ##

    from .iterators import (  # noqa
        PeekIterator,
        PrefetchIterator,
        ProxyIterator,
        RetainIterator,
    )

    from .recipes import (  # noqa
        sliding_window,
    )

    from .tools import (  # noqa
        expand_indexed_pairs,
        merge_on,
        take,
        unzip,
    )

    from . import transforms as tf  # noqa

    from .unique import (  # noqa
        UniqueItem,
        UniqueIterator,
        UniqueStats,
    )



# ruff: noqa: UP006 UP007 UP045 UP046
# @omlish-lite
import abc
import typing as ta

from ..lite.abstract import Abstract
from .contexts import CaptureLoggingContext
from .contexts import CaptureLoggingContextImpl
from .contexts import LoggingExcInfoArg
from .levels import LogLevel
from .levels import NamedLogLevel


T = ta.TypeVar('T')


LoggingMsgFn = ta.Callable[[], ta.Union[str, tuple]]  # ta.TypeAlias


##


class AnyLogger(Abstract, ta.Generic[T]):
    def is_enabled_for(self, level: LogLevel) -> bool:
        return level >= self.get_effective_level()

    @abc.abstractmethod
    def get_effective_level(self) -> LogLevel:
        raise NotImplementedError

    #

    @ta.final
    def isEnabledFor(self, level: LogLevel) -> bool:  # noqa
        return self.is_enabled_for(level)

    @ta.final
    def getEffectiveLevel(self) -> LogLevel:  # noqa
        return self.get_effective_level()

    ##

    @ta.overload
    def log(self, level: LogLevel, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def log(self, level: LogLevel, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def log(self, level: LogLevel, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def log(self, level: LogLevel, *args, **kwargs):
        return self._log(CaptureLoggingContextImpl(level, stack_offset=1), *args, **kwargs)

    #

    @ta.overload
    def debug(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def debug(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def debug(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def debug(self, *args, **kwargs):
        return self._log(CaptureLoggingContextImpl(NamedLogLevel.DEBUG, stack_offset=1), *args, **kwargs)

    #

    @ta.overload
    def info(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def info(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def info(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def info(self, *args, **kwargs):
        return self._log(CaptureLoggingContextImpl(NamedLogLevel.INFO, stack_offset=1), *args, **kwargs)

    #

    @ta.overload
    def warning(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def warning(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def warning(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def warning(self, *args, **kwargs):
        return self._log(CaptureLoggingContextImpl(NamedLogLevel.WARNING, stack_offset=1), *args, **kwargs)

    #

    @ta.overload
    def error(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def error(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def error(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def error(self, *args, **kwargs):
        return self._log(CaptureLoggingContextImpl(NamedLogLevel.ERROR, stack_offset=1), *args, **kwargs)

    #

    @ta.overload
    def exception(self, msg: str, *args: ta.Any, exc_info: LoggingExcInfoArg = True, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def exception(self, msg: ta.Tuple[ta.Any, ...], *, exc_info: LoggingExcInfoArg = True, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def exception(self, msg_fn: LoggingMsgFn, *, exc_info: LoggingExcInfoArg = True, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def exception(self, *args, exc_info: LoggingExcInfoArg = True, **kwargs):
        return self._log(CaptureLoggingContextImpl(NamedLogLevel.ERROR, exc_info=exc_info, stack_offset=1), *args, **kwargs)  # noqa

    #

    @ta.overload
    def critical(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def critical(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def critical(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def critical(self, *args, **kwargs):
        return self._log(CaptureLoggingContextImpl(NamedLogLevel.CRITICAL, stack_offset=1), *args, **kwargs)

    ##

    @abc.abstractmethod
    def _log(self, ctx: CaptureLoggingContext, msg: ta.Union[str, tuple, LoggingMsgFn], *args: ta.Any, **kwargs: ta.Any) -> T:  # noqa
        raise NotImplementedError


class Logger(AnyLogger[None], Abstract):
    @abc.abstractmethod
    def _log(self, ctx: CaptureLoggingContext, msg: ta.Union[str, tuple, LoggingMsgFn], *args: ta.Any, **kwargs: ta.Any) -> None:  # noqa
        raise NotImplementedError


class AsyncLogger(AnyLogger[ta.Awaitable[None]], Abstract):
    @abc.abstractmethod
    def _log(self, ctx: CaptureLoggingContext, msg: ta.Union[str, tuple, LoggingMsgFn], *args: ta.Any, **kwargs: ta.Any) -> ta.Awaitable[None]:  # noqa
        raise NotImplementedError


##


class AnyNopLogger(AnyLogger[T], Abstract):
    @ta.final
    def get_effective_level(self) -> LogLevel:
        return -999


@ta.final
class NopLogger(AnyNopLogger[None], Logger):
    def _log(self, ctx: CaptureLoggingContext, msg: ta.Union[str, tuple, LoggingMsgFn], *args: ta.Any, **kwargs: ta.Any) -> None:  # noqa
        pass


@ta.final
class AsyncNopLogger(AnyNopLogger[ta.Awaitable[None]], AsyncLogger):
    async def _log(self, ctx: CaptureLoggingContext, msg: ta.Union[str, tuple, LoggingMsgFn], *args: ta.Any, **kwargs: ta.Any) -> None:  # noqa
        pass



"""
TODO:
 - if already imported just return?
  - no, need sub-imports..
 - seal on first use? or just per module? can't seal roots and still be usable
  - only if not hasattr?
 - audit for deadlock risk - does importlib._bootstrap do it for us? do we need a global _ProxyImporter lock? would only
   happen on reification
 - ProxyImportError
 - detect import reification in own module body - user is failing to properly 'hands-off' lazy import
  - bonus points detect when done specifically for a type annotation

See:
 - https://peps.python.org/pep-0810/
  - https://github.com/LazyImportsCabal/cpython/tree/lazy
  - https://developers.facebook.com/blog/post/2022/06/15/python-lazy-imports-with-cinder/
  - https://engineering.fb.com/2024/01/18/developer-tools/lazy-imports-cinder-machine-learning-meta/
  - https://www.hudsonrivertrading.com/hrtbeat/inside-hrts-python-fork/
  - https://bugreports.qt.io/browse/PYSIDE-2404
 - https://scientific-python.org/specs/spec-0001/
 - https://github.com/scientific-python/lazy-loader
"""
import functools
import importlib.util
import threading
import types
import typing as ta

from ..lazyglobals import LazyGlobals
from .capture import ImportCapture
from .capture import _new_import_capture_hook


_ProxyImporterModuleAttr: ta.TypeAlias = ta.Literal[
    'child',  # 'outranks' proxy_attr - all child attrs must be proxy_attrs but not vice versa
    'proxy_attr',
    'pending_child',
    'pending_attr',
]


##


class _ProxyImporter:
    def __init__(
            self,
            *,
            owner_globals: ta.MutableMapping[str, ta.Any] | None = None,
    ) -> None:
        super().__init__()

        self._owner_globals = owner_globals

        self._owner_name: str | None = owner_globals.get('__name__') if owner_globals else None

        # NOTE: Import machinery may be reentrant for things like gc ops and signal handling:
        # TODO: audit for reentrancy this lol
        # https://github.com/python/cpython/blob/72f25a8d9a5673d39c107cf522465a566b979ed5/Lib/importlib/_bootstrap.py#L233-L237  # noqa
        self._lock = threading.RLock()

        self._modules_by_name: dict[str, _ProxyImporter._Module] = {}
        self._modules_by_proxy_obj: dict[types.ModuleType, _ProxyImporter._Module] = {}

    class _Module:
        def __init__(
                self,
                name: str,
                getattr_handler: ta.Callable[['_ProxyImporter._Module', str], ta.Any],
                *,
                parent: ta.Optional['_ProxyImporter._Module'] = None,
        ) -> None:
            super().__init__()

            self.name = name
            self.parent = parent

            self.base_name = name.rpartition('.')[2]
            self.root: _ProxyImporter._Module = parent.root if parent is not None else self  # noqa

            self.children: dict[str, _ProxyImporter._Module] = {}
            self.descendants: set[_ProxyImporter._Module] = set()

            self.proxy_obj = types.ModuleType(f'<{self.__class__.__qualname__}: {name}>')
            self.proxy_obj.__file__ = None
            self.proxy_obj.__getattr__ = functools.partial(getattr_handler, self)  # type: ignore[method-assign]

            self.pending_children: set[str] = set()
            self.pending_attrs: set[str] = set()

        real_obj: types.ModuleType | None = None

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}<{self.name}{"!" if self.real_obj is not None else ""}>'

        def find_attr(self, attr: str) -> _ProxyImporterModuleAttr | None:
            is_child = attr in self.children
            is_proxy_attr = attr in self.proxy_obj.__dict__
            is_pending_child = attr in self.pending_children
            is_pending_attr = attr in self.pending_attrs

            if is_child:
                if (
                        not is_proxy_attr or
                        is_pending_child or
                        is_pending_attr
                ):
                    raise RuntimeError
                return 'child'

            elif is_proxy_attr:
                if (
                        is_pending_child or
                        is_pending_attr
                ):
                    raise RuntimeError
                return 'proxy_attr'

            elif is_pending_child:
                if (
                        is_child or
                        is_proxy_attr or
                        is_pending_attr
                ):
                    raise RuntimeError
                return 'pending_child'

            elif is_pending_attr:
                if (
                        is_child or
                        is_proxy_attr or
                        is_pending_child
                ):
                    raise RuntimeError
                return 'pending_attr'

            else:
                return None

    #

    def _get_or_make_module_locked(self, name: str) -> _Module:
        try:
            return self._modules_by_name[name]
        except KeyError:
            pass

        parent: _ProxyImporter._Module | None = None
        if '.' in name:
            rest, _, attr = name.rpartition('.')
            parent = self._get_or_make_module_locked(rest)

            fa = parent.find_attr(attr)
            if not (fa == 'pending_child' or fa is None):
                raise RuntimeError

            if (ro := parent.real_obj) is not None and attr not in ro.__dict__:
                raise NotImplementedError

        module = self._modules_by_name[name] = _ProxyImporter._Module(
            name,
            self._handle_module_getattr,
        )

        self._modules_by_name[name] = module
        self._modules_by_proxy_obj[module.proxy_obj] = module

        if parent is not None:
            parent.pending_children.discard(module.base_name)
            parent.children[module.base_name] = module
            setattr(parent.proxy_obj, module.base_name, module.proxy_obj)
            parent.root.descendants.add(module)

        return module

    def _extend_module_locked(
            self,
            module: _Module,
            *,
            children: ta.Iterable[str] | None = None,
            attrs: ta.Iterable[str] | None = None,
    ) -> None:
        for l in (children, attrs):
            for n in l or ():
                if n in module.proxy_obj.__dict__:
                    raise NotImplementedError

                if (ro := module.real_obj) is not None and n in ro.__dict__:
                    raise NotImplementedError

        for n in children or ():
            fa = module.find_attr(n)
            if not (fa == 'pending_child' or fa is None):
                raise RuntimeError

        for n in attrs or ():
            fa = module.find_attr(n)
            if not (fa == 'pending_attr' or fa is None):
                raise RuntimeError

        #

        if children:
            module.pending_children.update(n for n in children if n not in module.children)
        if attrs:
            module.pending_attrs.update(attrs)

    def _retrieve_from_module_locked(self, module: _Module, attr: str) -> ta.Any:
        fa = module.find_attr(attr)

        if fa == 'child' or fa == 'proxy_attr':
            return module.proxy_obj.__dict__[attr]

        val: ta.Any

        if fa == 'pending_child':
            if module.name == self._owner_name:
                val = importlib.import_module(f'{module.name}.{attr}')

            else:
                mod = __import__(
                    module.name,
                    self._owner_globals or {},
                    {},
                    [attr],
                    0,
                )

                val = getattr(mod, attr)

            module.pending_children.remove(attr)

        elif fa == 'pending_attr' or fa is None:
            if module.name == self._owner_name:
                raise NotImplementedError

            if (ro := module.real_obj) is None:
                ro = module.real_obj = importlib.import_module(module.name)

            val = getattr(ro, attr)

            if fa == 'pending_attr':
                module.pending_attrs.remove(attr)

        else:
            raise TypeError(fa)

        setattr(module.proxy_obj, attr, val)
        return val

    #

    def _handle_module_getattr(self, module: _Module, attr: str) -> ta.Any:
        with self._lock:
            return self._retrieve_from_module_locked(module, attr)

    def add(
            self,
            module: str,
            *,
            children: ta.Iterable[str] | None = None,
            attrs: ta.Iterable[str] | None = None,
    ) -> types.ModuleType:
        if isinstance(children, str):
            raise TypeError(children)

        # Leaf modules get collapsed into parents' pending_children
        if not children and not attrs and '.' in module:
            module, _, child = module.rpartition('.')
            children = [child]

        with self._lock:
            m = self._get_or_make_module_locked(module)

            if children or attrs:
                self._extend_module_locked(
                    m,
                    children=children,
                    attrs=attrs,
                )

        return m.proxy_obj

    def get_module(self, name: str) -> types.ModuleType:
        try:
            return self._modules_by_name[name].proxy_obj
        except KeyError:
            pass

        with self._lock:
            return self._get_or_make_module_locked(name).proxy_obj

    def lookup(self, spec: str) -> ta.Any:
        if '.' not in spec:
            return self._modules_by_name[spec].proxy_obj

        try:
            module = self._modules_by_name[spec]
        except KeyError:
            pass
        else:
            return module.proxy_obj

        rest, _, attr = spec.rpartition('.')
        module = self._modules_by_name[rest]
        return getattr(module.proxy_obj, attr)


#


_MODULE_PROXY_IMPORTER_GLOBAL_NAME = '__proxy_importer__'


def _get_module_proxy_importer(mod_globals: ta.MutableMapping[str, ta.Any]) -> _ProxyImporter:
    """Assumed to only be called in a module body - no locking is done."""

    pi: _ProxyImporter
    try:
        pi = mod_globals[_MODULE_PROXY_IMPORTER_GLOBAL_NAME]

    except KeyError:
        pi = _ProxyImporter(
            owner_globals=mod_globals,
        )
        mod_globals[_MODULE_PROXY_IMPORTER_GLOBAL_NAME] = pi

    else:
        if pi.__class__ is not _ProxyImporter:
            raise TypeError(pi)

        if pi._owner_globals is not mod_globals:  # noqa
            raise RuntimeError

    return pi


##


def proxy_import(
        spec: str,
        package: str | None = None,
        extras: ta.Iterable[str] | None = None,
) -> types.ModuleType:
    """'Legacy' proxy import mechanism."""

    if isinstance(extras, str):
        raise TypeError(extras)

    omod = None

    def __getattr__(att):  # noqa
        nonlocal omod
        if omod is None:
            omod = importlib.import_module(spec, package=package)
            if extras:
                for x in extras:
                    importlib.import_module(f'{spec}.{x}', package=package)
        return getattr(omod, att)

    lmod = types.ModuleType(spec)
    lmod.__getattr__ = __getattr__  # type: ignore[method-assign]
    return lmod


#


def proxy_init(
        init_globals: ta.MutableMapping[str, ta.Any],
        spec: str,
        attrs: ta.Iterable[str | tuple[str | None, str | None] | None] | None = None,
) -> None:
    name = importlib.util.resolve_name(
        spec,
        package=init_globals['__package__'] if spec.startswith('.') else None,
    )

    #

    if isinstance(attrs, str):
        raise TypeError(attrs)

    if attrs is None:
        attrs = [None]

    whole_attr = spec.split('.')[-1]
    al: list[tuple[str | None, str]] = []
    for attr in attrs:
        if attr is None:
            al.append((None, whole_attr))

        elif isinstance(attr, str):
            al.append((attr, attr))

        elif isinstance(attr, tuple):
            imp_attr, as_attr = attr
            if as_attr is None:
                if imp_attr is None:
                    as_attr = whole_attr
                else:
                    as_attr = imp_attr
            al.append((imp_attr, as_attr))

        else:
            raise TypeError(attr)

    #

    pi = _get_module_proxy_importer(init_globals)
    lg = LazyGlobals.install(init_globals)

    pi.add(
        name,
        children=[r for l, r in al if r is not None],
    )

    for imp_attr, as_attr in al:
        lg.set_fn(as_attr, functools.partial(pi.lookup, name if imp_attr is None else f'{name}.{imp_attr}'))


##


class _AutoProxy:
    def __init__(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],
            *,
            disable: bool = False,
            eager: bool = False,

            unreferenced_callback: ta.Callable[[ta.Sequence[str]], None] | None = None,
            raise_unreferenced: bool = False,

            update_exports: bool = False,

            _stack_offset: int = 0,
            _capture_impl: str | None = None,
    ) -> None:
        super().__init__()

        self._mod_globals = mod_globals

        self._disabled = disable
        self._eager = eager

        self._unreferenced_callback = unreferenced_callback
        self._raise_unreferenced = raise_unreferenced

        self._update_exports = update_exports

        self._ic = ImportCapture(
            mod_globals,
            _hook=_new_import_capture_hook(
                mod_globals,
                stack_offset=_stack_offset + 1,
                capture_impl=_capture_impl,
            ),
            disable=disable,
        )
        self._icc: ta.Any = None

    def __enter__(self) -> ImportCapture:
        if self._icc is not None:
            raise RuntimeError

        self._icc = self._ic.capture(
            unreferenced_callback=self._unreferenced_callback,
            raise_unreferenced=self._raise_unreferenced,
        )

        return self._icc.__enter__()  # noqa

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._icc is None:
            raise RuntimeError

        self._icc.__exit__(exc_type, exc_val, exc_tb)

        if not self._disabled and exc_type is None:
            self._install()

    # @abc.abstractmethod
    def _install(self) -> None:
        raise NotImplementedError


@ta.final
class _AutoProxyImport(_AutoProxy):
    def _install(self) -> None:
        cap = self._ic.captured

        for cm in cap.modules.values():
            if cm.attrs:
                raise RuntimeError

        pi = _get_module_proxy_importer(self._mod_globals)

        for cm in cap.modules.values():
            pi.add(
                cm.name,
                children=cm.children,
            )

        for ci in cap.imports.values():
            pm = pi.get_module(ci.module.name)
            for a in ci.as_ or ():
                self._mod_globals[a] = pm

        if self._eager:
            for ci in cap.imports.values():
                pi.lookup(ci.module.name)

        if self._update_exports:
            self._ic.update_exports()


@ta.final
class _AutoProxyInit(_AutoProxy):
    def _install(self) -> None:
        cap = self._ic.captured

        pi = _get_module_proxy_importer(self._mod_globals)
        lg = LazyGlobals.install(self._mod_globals)

        for cm in cap.modules.values():
            pi.add(
                cm.name,
                children=cm.children,
                attrs=cm.attrs,
            )

        for ci in cap.imports.values():
            for a in ci.as_ or ():
                lg.set_fn(a, functools.partial(pi.lookup, ci.module.name))
            for sa, da in ci.attrs or ():
                lg.set_fn(da, functools.partial(pi.lookup, f'{ci.module.name}.{sa}'))

        if self._eager:
            for a in cap.attrs:
                lg.get(a)

        if self._update_exports:
            self._ic.update_exports()


auto_proxy_import = _AutoProxyImport
auto_proxy_init = _AutoProxyInit



import typing as ta

from .. import lang
from .keys import Key
from .keys import as_key


if ta.TYPE_CHECKING:
    from .impl import inspect as _inspect
else:
    _inspect = lang.proxy_import('.impl.inspect', __package__)


T = ta.TypeVar('T')


##


class Kwarg(ta.NamedTuple):
    name: str
    key: Key
    has_default: bool

    @classmethod
    def of(
            cls,
            name: str,
            key: Key,
            *,
            has_default: bool = False,
    ) -> 'Kwarg':
        return cls(
            name,
            key,
            has_default,
        )


class KwargsTarget(ta.NamedTuple):
    obj: ta.Any
    kwargs: ta.Sequence[Kwarg]

    @classmethod
    def of(
            cls,
            obj: ta.Any,
            *kws: Kwarg,
            **kwargs: tuple[Key, bool] | Key | ta.Any,
    ) -> 'KwargsTarget':
        kw_kwargs: list[Kwarg] = []
        for n, v in kwargs.items():
            if isinstance(v, tuple):
                kw_k, kw_hd = v
                kw_kwargs.append(Kwarg.of(n, kw_k, has_default=kw_hd))
            else:
                kw_kwargs.append(Kwarg.of(n, as_key(v)))

        return cls(
            obj,
            [*kws, *kw_kwargs],
        )


def tag(obj: T, **kwargs: ta.Any) -> T:
    return _inspect.tag(obj, **kwargs)


def build_kwargs_target(obj: ta.Any, **kwargs: ta.Any) -> KwargsTarget:
    return _inspect.build_kwargs_target(obj, **kwargs)



import typing as ta

from .. import codecs
from .. import reflect as rfl


ObjectCodecT = ta.TypeVar('ObjectCodecT', bound='ObjectCodec')


##


class ObjectCodec(codecs.Codec):
    pass


def make_object_codec(
        cls: type[ObjectCodecT],
        name: str,
        dumps: ta.Callable,
        loads: ta.Callable,
        *,
        input: rfl.Type,  # noqa
        output: rfl.Type = rfl.type_(ta.Any),
        aliases: ta.Sequence[str] | None = None,
) -> ObjectCodecT:
    return cls(
        name=name,
        aliases=aliases,

        input=input,
        output=output,

        new=lambda: codecs.FnPairEagerCodec.of(dumps, loads),
    )


##


class BytesObjectCodec(ObjectCodec):
    pass


def make_bytes_object_codec(
        name: str,
        dumps: ta.Callable[[ta.Any], bytes],
        loads: ta.Callable[[bytes], ta.Any],
        **kwargs: ta.Any,
) -> BytesObjectCodec:
    return make_object_codec(
        BytesObjectCodec,
        name,
        dumps,
        loads,
        input=bytes,
        **kwargs,
    )


##


class StrObjectCodec(ObjectCodec):
    pass


def make_str_object_codec(
        name: str,
        dumps: ta.Callable[[ta.Any], str],
        loads: ta.Callable[[str], ta.Any],
        **kwargs: ta.Any,
) -> StrObjectCodec:
    return make_object_codec(
        StrObjectCodec,
        name,
        dumps,
        loads,
        input=str,
        **kwargs,
    )


##


def make_object_lazy_loaded_codec(
        module: str,
        attr: str,
        codec: ObjectCodec,
) -> codecs.LazyLoadedCodec:
    return codecs.LazyLoadedCodec.new(
        module,
        attr,
        codec,
    )



from .descriptor import (  # noqa
    Ignore,
    Scope,
    cache,
    ignore,
)

from .impl import (  # noqa
    LFU,
    LRI,
    LRU,
    new_cache,
)

from .types import (  # noqa
    Cache,
    Eviction,
    OverweightError,
)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta


##


@dc.dataclass(frozen=True)
class ManifestOrigin:
    module: str  # Always starts with exactly one '.'
    attr: ta.Optional[str]  # None if inline

    file: str
    line: int


@dc.dataclass(frozen=True)
class Manifest(ManifestOrigin):
    value: ta.Any  # [{class_key: value_dct}], where class_key is of the form `!.foo.bar.Class` or `!baz.quz.Class`



# @omlish-lite
import abc
import typing as ta

from ..lite.abstract import Abstract


T = ta.TypeVar('T')


##


class AsyncBufferRelay(Abstract, ta.Generic[T]):
    @abc.abstractmethod
    def push(self, *vs: T) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def swap(self) -> ta.Sequence[T]:
        raise NotImplementedError

    @abc.abstractmethod
    async def wait(self) -> None:
        raise NotImplementedError



import typing as ta

from ..impl.api.classes.decorator import dataclass
from ..impl.concerns.frozen import unchecked_frozen_base
from .meta import DataMeta
from .specs import get_metaclass_spec


T = ta.TypeVar('T')


##


# @ta.dataclass_transform(field_specifiers=(field,))  # FIXME: ctor
@unchecked_frozen_base
class Data(
    eq=False,
    order=False,
    allow_redundant_decorator=True,
    confer=frozenset([
        'confer',
        'final_subclasses',
        'allow_redundant_decorator',
    ]),
    metaclass=DataMeta,
):
    def __init__(self, *args, **kwargs):
        # Typechecking barrier
        super().__init__(*args, **kwargs)

    def __init_subclass__(cls, **kwargs):
        # Typechecking barrier
        super().__init_subclass__(**kwargs)


#


class Frozen(
    Data,
    frozen=True,
    eq=False,
    order=False,
    confer=frozenset([
        *get_metaclass_spec(Data).confer,
        'frozen',
        'reorder',
        'cache_hash',
        'override',
    ]),
):
    pass


#


class Case(
    Frozen,
    abstract=True,
    override=True,
    final_subclasses=True,
    abstract_immediate_subclasses=True,
):
    pass


#


@dataclass(frozen=True)
class Box(
    Frozen,
    ta.Generic[T],
    generic_init=True,
    terse_repr=True,
    confer=frozenset([
        *get_metaclass_spec(Frozen).confer,
        'generic_init',
        'terse_repr',
    ]),
):
    v: T



import abc
import typing as ta

from .abstract import Abstract
from .restrict import NotInstantiable


V = ta.TypeVar('V')


##


class _NOT_SET:  # noqa
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError


NamespaceCheckValue: ta.TypeAlias = type | tuple[type, ...] | ta.Callable[[ta.Any], bool]


class GenericNamespaceMeta(abc.ABCMeta, ta.Generic[V]):
    __namespace_check_values__: NamespaceCheckValue | None = None
    __namespace_case_insensitive__: bool = False

    def __init_subclass__(
            mcls,
            *,
            check_values: NamespaceCheckValue | None | type[_NOT_SET] = _NOT_SET,
            case_insensitive: bool | type[_NOT_SET] = _NOT_SET,
            **kwargs,
    ):
        super().__init_subclass__(**kwargs)

        if check_values is not _NOT_SET:
            mcls.__namespace_check_values__ = check_values
        if case_insensitive is not _NOT_SET:
            mcls.__namespace_case_insensitive__ = case_insensitive  # type: ignore

    __namespace_attrs__: ta.Mapping[str, V]
    __namespace_values__: ta.Mapping[str, V]

    def __new__(
            mcls,
            name,
            bases,
            namespace,
            *,
            check_values: NamespaceCheckValue | None | type[_NOT_SET] = _NOT_SET,
            case_insensitive: bool | type[_NOT_SET] = _NOT_SET,
            **kwargs,
    ):
        if bases:
            if NotInstantiable not in bases and not any(NotInstantiable in b.__mro__ for b in bases):
                if Abstract in bases:
                    # Must go before Abstract for mro because NotInstantiable is itself Abstract
                    ai = bases.index(Abstract)
                    bases = (*bases[:ai], NotInstantiable, *bases[ai:])
                else:
                    bases += (NotInstantiable,)

        if check_values is _NOT_SET:
            check_values = mcls.__namespace_check_values__
        if case_insensitive is _NOT_SET:
            case_insensitive = mcls.__namespace_case_insensitive__

        cls = super().__new__(
            mcls,
            name,
            bases,
            namespace,
            **kwargs,
        )

        cls.__namespace_check_values__ = check_values
        cls.__namespace_case_insensitive__ = case_insensitive  # type: ignore

        if isinstance(check_values, (type, tuple)):
            cvt = check_values

            def _cv(o):
                if not isinstance(o, cvt):
                    raise TypeError(o, cvt)
                return True

            check_values = _cv

        a_dct: dict[str, V] = {}
        v_dct: dict[str, V] = {}
        for b_cls in cls.__mro__[-2::-1]:
            # FIXME: must list() to avoid `RuntimeError: dictionary changed size during iteration`
            for att in list(b_cls.__dict__):
                if att.startswith('_') or att in a_dct:
                    continue

                name = att
                if case_insensitive:
                    name = att.lower()
                    if name in v_dct:
                        raise NameError(f'Ambiguous case-insensitive namespace attr: {name!r}, {att!r}')

                obj = getattr(cls, att)
                if check_values is not None and not check_values(obj):
                    raise ValueError(obj)

                a_dct[att] = obj
                v_dct[name] = obj

        cls.__namespace_attrs__ = a_dct
        cls.__namespace_values__ = v_dct

        return cls

    def __iter__(cls) -> ta.Iterator[tuple[str, V]]:
        return iter(cls.__namespace_attrs__.items())

    def __contains__(self, n: str) -> bool:
        return n in self.__namespace_values__

    def __getitem__(cls, n: str) -> V:
        if cls.__namespace_case_insensitive__:
            n = n.lower()
        return cls.__namespace_values__[n]


class NamespaceMeta(GenericNamespaceMeta[ta.Any]):
    pass


class Namespace(metaclass=NamespaceMeta):
    pass



"""
This module solely exists for import convenience - tools will tend to resolve the imported 'property' as if it were from
builtins and thus not distinguish it from a normal property.

    from omlish import cached

    class C:
        @cached.property
        def p(self) -> str: ...

"""
from .lang.cached.function import cached_function as _cached_function
from .lang.cached.property import cached_property as _cached_property


function = _cached_function

property = property  # noqa

globals()['property'] = _cached_property  # noqa



"""
TODO:
 - cext
"""
import array
import codecs
import typing as ta


##


def decode_to_list(
        raw: bytes,
        encoding: str = 'utf-8',
        *,
        errors: str = 'strict',
) -> list[str]:
    dec = codecs.getincrementaldecoder(encoding)(errors)
    end = len(raw) - 1
    return [dec.decode(bytes([b]), final=i == end) for i, b in enumerate(raw)]


#


class DecodedStringIndex(ta.NamedTuple):
    byte_offsets: ta.Sequence[int]


def decode_indexed(
        raw: bytes,
        encoding: str = 'utf-8',
        *,
        errors: str = 'strict',
) -> tuple[str, DecodedStringIndex]:
    dec_lst = decode_to_list(
        raw,
        encoding,
        errors=errors,
    )

    dec_s = ''.join(dec_lst)

    bo_arr_len = len(dec_s) + 1
    if bo_arr_len < 2**8:
        fmt = 'B'
    elif bo_arr_len < 2**16:
        fmt = 'H'
    elif bo_arr_len < 2**32:
        fmt = 'L'
    else:
        fmt = 'Q'

    bo_arr = array.array(fmt, [0]) * bo_arr_len
    so = 0
    lbo = 0
    for bo, s in enumerate(dec_lst):
        if s:
            bo_arr[so] = lbo
            so += 1
            lbo = bo + 1

    if so != len(dec_s):
        raise RuntimeError

    bo_arr[len(dec_s)] = len(raw)

    dsi = DecodedStringIndex(
        bo_arr,
    )

    return (dec_s, dsi)



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
import io
import typing as ta

from ..lite.attrops import attr_repr
from ..lite.check import check


##


class DelimitingBuffer:
    """
    https://github.com/python-trio/trio/issues/796 :|

    FIXME: when given overlapping delimiters like [b'\r', b'\r\n'], *should* refuse to output a line ending in '\r'
      without knowing it will not be followed by '\n'. does not currently do this - currently only picks longest
      delimiter present in the buffer. does this need a prefix-trie? is this borderline parsing?
    """

    #

    class Error(Exception):
        def __init__(self, buffer: 'DelimitingBuffer') -> None:
            super().__init__(buffer)

            self.buffer = buffer

        def __repr__(self) -> str:
            return attr_repr(self, 'buffer')

    class ClosedError(Error):
        pass

    #

    DEFAULT_DELIMITERS: bytes = b'\n'

    def __init__(
            self,
            delimiters: ta.Iterable[ta.Union[int, bytes]] = DEFAULT_DELIMITERS,
            *,
            keep_ends: bool = False,
            max_size: ta.Optional[int] = None,
    ) -> None:
        super().__init__()

        ds: ta.Set[bytes] = set()
        for d in delimiters:
            if isinstance(d, int):
                d = bytes([d])
            ds.add(check.isinstance(d, bytes))

        self._delimiters: ta.FrozenSet[bytes] = frozenset(ds)
        self._keep_ends = keep_ends
        self._max_size = max_size

        self._buf: ta.Optional[io.BytesIO] = io.BytesIO()

        ddl = {}
        dl = sorted(self._delimiters, key=lambda d: -len(d))
        for i, d in enumerate(dl):
            for j, d2 in enumerate(dl):
                if len(d2) < len(d):
                    break
                if i == j or not d2.startswith(d):
                    continue
                ddl[d] = len(d2)
                break
        self._delimiter_disambiguation_lens: ta.Dict[bytes, int] = ddl

    #

    @property
    def is_closed(self) -> bool:
        return self._buf is None

    def tell(self) -> int:
        if (buf := self._buf) is None:
            raise self.ClosedError(self)
        return buf.tell()

    def peek(self) -> bytes:
        if (buf := self._buf) is None:
            raise self.ClosedError(self)
        return buf.getvalue()

    def _find_delim(self, data: ta.Union[bytes, bytearray], i: int) -> ta.Optional[ta.Tuple[int, int]]:
        rp = None  # type: int | None
        rl = None  # type: int | None
        rd = None  # type: bytes | None

        for d in self._delimiters:
            if (p := data.find(d, i)) < 0:
                continue

            dl = len(d)

            if rp is None or p < rp:
                rp, rl, rd = p, dl, d
            elif rp == p and rl < dl:  # type: ignore
                rl, rd = dl, d  # noqa

        if rp is None:
            return None

        # FIXME:
        # if (ddl := self._delimiter_disambiguation_lens.get(rd)) is not None:
        #     raise NotImplementedError

        return rp, rl  # type: ignore

    def _append_and_reset(self, chunk: bytes) -> bytes:
        buf = check.not_none(self._buf)
        if not buf.tell():
            return chunk

        buf.write(chunk)
        ret = buf.getvalue()
        buf.seek(0)
        buf.truncate()
        return ret

    class Incomplete(ta.NamedTuple):
        b: bytes

    def feed(self, data: ta.Union[bytes, bytearray]) -> ta.Generator[ta.Union[bytes, Incomplete], None, None]:
        if (buf := self._buf) is None:
            raise self.ClosedError(self)

        if not data:
            self._buf = None

            if buf.tell():
                yield self.Incomplete(buf.getvalue())

            return

        l = len(data)
        i = 0
        while i < l:
            if (pt := self._find_delim(data, i)) is None:
                break

            p, pl = pt
            n = p + pl
            if self._keep_ends:
                p = n

            yield self._append_and_reset(data[i:p])

            i = n

        if i >= l:
            return

        if self._max_size is None:
            buf.write(data[i:])
            return

        while i < l:
            remaining_data_len = l - i
            remaining_buf_capacity = self._max_size - buf.tell()

            if remaining_data_len < remaining_buf_capacity:
                buf.write(data[i:])
                return

            p = i + remaining_buf_capacity
            yield self.Incomplete(self._append_and_reset(data[i:p]))
            i = p


##


class ReadableListBuffer:
    # FIXME: merge with PrependableGeneratorReader

    def __init__(self) -> None:
        super().__init__()

        self._lst: list[bytes] = []

    def __len__(self) -> int:
        return sum(map(len, self._lst))

    def feed(self, d: bytes) -> None:
        if d:
            self._lst.append(d)

    def _chop(self, i: int, e: int) -> bytes:
        lst = self._lst
        d = lst[i]

        o = b''.join([
            *lst[:i],
            d[:e],
        ])

        self._lst = [
            *([d[e:]] if e < len(d) else []),
            *lst[i + 1:],
        ]

        return o

    def read(self, n: ta.Optional[int] = None) -> ta.Optional[bytes]:
        if n is None:
            if not self._lst:
                return b''

            o = b''.join(self._lst)
            self._lst = []
            return o

        if not (lst := self._lst):
            return None

        c = 0
        for i, d in enumerate(lst):
            r = n - c
            if (l := len(d)) >= r:
                return self._chop(i, r)
            c += l

        return None

    def read_exact(self, sz: int) -> bytes:
        d = self.read(sz)
        if d is None or len(d) != sz:
            raise EOFError(f'ReadableListBuffer got {"no" if d is None else len(d)}, expected {sz}')
        return d

    def read_until_(self, delim: bytes = b'\n', start_buffer: int = 0) -> ta.Union[bytes, int]:
        if not (lst := self._lst):
            return 0

        i = start_buffer
        while i < len(lst):
            if (p := lst[i].find(delim)) >= 0:
                return self._chop(i, p + len(delim))
            i += 1

        return i

    def read_until(self, delim: bytes = b'\n') -> ta.Optional[bytes]:
        r = self.read_until_(delim)
        return r if isinstance(r, bytes) else None


##


class IncrementalWriteBuffer:
    def __init__(
            self,
            data: bytes,
            *,
            write_size: int = 0x10000,
    ) -> None:
        super().__init__()

        check.not_empty(data)
        self._len = len(data)
        self._write_size = write_size

        self._lst = [
            data[i:i + write_size]
            for i in range(0, len(data), write_size)
        ]
        self._pos = 0

    @property
    def rem(self) -> int:
        return self._len - self._pos

    def write(self, fn: ta.Callable[[bytes], int]) -> int:
        lst = check.not_empty(self._lst)

        t = 0
        for i, d in enumerate(lst):  # noqa
            n = fn(check.not_empty(d))
            if not n:
                break
            t += n

        if t:
            self._lst = [
                *([d[n:]] if n < len(d) else []),
                *lst[i + 1:],
            ]
            self._pos += t

        return t



"""
TODO:
 - typed_lambda is really kinda just 'annotate'
 - TypedLambda / TypedPartial classes, picklable, reprs
  - probably need to gen types per inst
 - typed_factory
"""
import inspect
import typing as ta

from ..lite.wrappers import update_wrapper_no_annotations


Ty = ta.TypeVar('Ty', bound=type)

T = ta.TypeVar('T')
T_co = ta.TypeVar('T_co', covariant=True)
T_contra = ta.TypeVar('T_contra', contravariant=True)

# FIXME: remove? ducktyped by mypy (with memoryview)
BytesLike: ta.TypeAlias = bytes | bytearray


##


class static_check_isinstance(ta.Generic[T]):  # noqa
    def __init__(self, *o: T) -> None:
        pass

    def __call__(self, o: T) -> T:
        return o


class static_check_issubclass(ta.Generic[T]):  # noqa
    def __init__(self, *t: type[T]) -> None:
        pass

    def __call__(self, t: type[T]) -> type[T]:
        return t


def copy_type(o: T) -> ta.Callable[[ta.Any], T]:
    """https://github.com/python/typing/issues/769#issuecomment-903760354"""

    return lambda x: x


##


def protocol_check(proto: type) -> ta.Callable[[Ty], Ty]:
    def inner(cls):
        if not issubclass(cls, proto):
            raise TypeError(cls)
        return cls
    return inner


##


_MISSING = object()


def typed_lambda(ret=_MISSING, **kw):  # noqa
    def inner(fn):
        ns = {}
        ns['__fn'] = fn

        proto = ['def __lam(']
        call = ['return __fn(']

        pkw = dict(kw)
        for i, (n, t) in enumerate(pkw.items()):
            if i:
                call.append(', ')
            else:
                proto.append('*')

            ns['__ann_' + n] = t
            proto.append(f', {n}: __ann_{n}')
            call.append(f'{n}={n}')

        proto.append(')')

        if ret is not _MISSING:
            ns['__ann_return'] = ret
            proto.append(f' -> __ann_return')

        proto.append(':')
        call.append(')')

        src = f'{"".join(proto)} {"".join(call)}'
        exec(src, ns)

        lam = update_wrapper_no_annotations(ns['__lam'], fn)
        lam.__signature__ = inspect.signature(lam, follow_wrapped=False)
        return lam

    for k in kw:
        if k.startswith('__'):
            raise NameError(k)

    return inner


def typed_partial(obj, **kw):  # noqa
    for k in kw:
        if k.startswith('__'):
            raise NameError(k)

    sig = inspect.signature(obj)

    inner = update_wrapper_no_annotations(lambda **lkw: obj(**lkw, **kw), obj)

    ret = (
        obj if isinstance(obj, type) else
        sig.return_annotation if sig.return_annotation is not inspect.Signature.empty else
        _MISSING
    )

    lam = typed_lambda(
        ret,
        **{
            n: p.annotation
            for n, p in sig.parameters.items()
            if n not in kw
            and p.annotation is not inspect.Signature.empty
        },
    )(inner)

    return update_wrapper_no_annotations(lam, obj)


##


class SequenceNotStr(ta.Protocol[T_co]):
    """
    https://github.com/python/mypy/issues/11001
    https://github.com/python/typing/issues/256#issuecomment-1442633430
    https://github.com/hauntsaninja/useful_types/blob/735ef9dd0b55b35b118ef630d5a0f3618ecedbff/useful_types/__init__.py#L285
    """

    @ta.overload
    def __getitem__(self, index: ta.SupportsIndex, /) -> T_co: ...

    @ta.overload
    def __getitem__(self, index: slice, /) -> ta.Sequence[T_co]: ...  # noqa

    def __contains__(self, value: object, /) -> bool: ...

    def __len__(self) -> int: ...

    def __iter__(self) -> ta.Iterator[T_co]: ...

    def index(self, value: ta.Any, start: int = 0, stop: int = ..., /) -> int: ...

    def count(self, value: ta.Any, /) -> int: ...

    def __reversed__(self) -> ta.Iterator[T_co]: ...



import collections
import dataclasses as dc
import sys
import typing as ta

from ..... import lang
from ....specs import CoerceFn
from ....specs import ReprFn
from ....specs import ValidateFn
from .metadata import _ExtraFieldParamsMetadata
from .metadata import extra_field_params


##


_IS_PY_3_14 = sys.version_info >= (3, 14)


def field(
        default=dc.MISSING,
        *,
        default_factory=dc.MISSING,
        init=True,
        repr=True,  # noqa
        hash=None,  # noqa
        compare=True,
        metadata=None,
        kw_only=dc.MISSING,
        doc=None,

        coerce: bool | CoerceFn | None = None,
        validate: ValidateFn | None = None,  # noqa
        check_type: bool | type | tuple[type | None, ...] | None = None,
        override: bool | None = None,
        repr_fn: ReprFn | None = None,
        repr_priority: int | None = None,
) -> ta.Any:
    efp = extra_field_params(**{
        **(metadata.get(_ExtraFieldParamsMetadata) or {} if metadata is not None else {}),
        **lang.opt_kw(
            coerce=coerce,
            validate=validate,
            check_type=check_type,
            override=override,
            repr_fn=repr_fn,
            repr_priority=repr_priority,
            doc=doc,
        ),
    })

    md: ta.Any = metadata
    if md is None:
        md = efp
    else:
        md = collections.ChainMap(efp, md)  # type: ignore[arg-type]

    return dc.field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=md,
        kw_only=kw_only,
        **(dict(doc=doc) if _IS_PY_3_14 else {}),
    )



# ruff: noqa: UP043 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from .asyncs import AsyncHttpClient
from .asyncs import AsyncStreamHttpResponse
from .base import HttpClientContext
from .base import HttpRequest
from .sync import HttpClient
from .sync import StreamHttpResponse


##


class ExecutorAsyncHttpClient(AsyncHttpClient):
    def __init__(
            self,
            run_in_executor: ta.Callable[..., ta.Awaitable],
            client: HttpClient,
    ) -> None:
        super().__init__()

        self._run_in_executor = run_in_executor
        self._client = client

    @dc.dataclass(frozen=True)
    class _StreamAdapter:
        owner: 'ExecutorAsyncHttpClient'
        resp: StreamHttpResponse

        async def read1(self, /, n: int = -1) -> bytes:
            return await self.owner._run_in_executor(self.resp.stream.read1, n)  # noqa

        async def close(self) -> None:
            return await self.owner._run_in_executor(self.resp.close)  # noqa

    async def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> AsyncStreamHttpResponse:
        resp: StreamHttpResponse = await self._run_in_executor(lambda: self._client.stream_request(req, context=ctx))
        return AsyncStreamHttpResponse(
            status=resp.status,
            headers=resp.headers,
            request=req,
            underlying=resp,
            **(dict(  # type: ignore
                stream=(adapter := self._StreamAdapter(self, resp)),
                _closer=adapter.close,
            ) if resp.has_data else {}),
        )



import contextlib


##


@contextlib.contextmanager
def assert_raises_star(et):
    num_caught = 0
    try:
        yield
    except* et as eg:  # noqa
        num_caught += 1
    assert num_caught > 0



# ruff: noqa: UP006 UP007 UP045
import typing as ta

from .handlers import FdioHandler
from .pollers import FdioPoller


##


class FdioManager:
    def __init__(
            self,
            poller: FdioPoller,
    ) -> None:
        super().__init__()

        self._poller = poller

        self._handlers: ta.Dict[int, FdioHandler] = {}  # Preserves insertion order

    def register(self, h: FdioHandler) -> None:
        if (hid := id(h)) in self._handlers:
            raise KeyError(h)
        self._handlers[hid] = h

    def unregister(self, h: FdioHandler) -> None:
        del self._handlers[id(h)]

    def poll(self, *, timeout: float = 1.) -> None:
        hs = list(self._handlers.values())
        rd = {h.fd(): h for h in hs if h.readable()}
        wd = {h.fd(): h for h in hs if h.writable()}

        self._poller.update(set(rd), set(wd))

        pr = self._poller.poll(timeout)

        for f in pr.r:
            if not (h := rd[f]).closed:
                h.on_readable()
        for f in pr.w:
            if not (h := wd[f]).closed:
                h.on_writable()

        hs = list(self._handlers.values())
        nh = {}
        for h in hs:
            if not h.closed:
                nh[id(h)] = h
        self._handlers = nh



# ruff: noqa: UP006 UP007 UP045
import typing as ta

from ..configs.formats import DEFAULT_CONFIG_FILE_LOADER
from ..configs.types import ConfigMap
from .marshal import OBJ_MARSHALER_MANAGER
from .marshal import ObjMarshalerManager


T = ta.TypeVar('T')


##


def load_config_file_obj(
        f: str,
        cls: ta.Type[T],
        *,
        prepare: ta.Union[
            ta.Callable[[ConfigMap], ConfigMap],
            ta.Iterable[ta.Callable[[ConfigMap], ConfigMap]],
        ] = (),
        msh: ObjMarshalerManager = OBJ_MARSHALER_MANAGER,
) -> T:
    config_data = DEFAULT_CONFIG_FILE_LOADER.load_file(f)

    config_dct = config_data.as_map()

    if prepare is not None:
        if isinstance(prepare, ta.Iterable):
            pfs = list(prepare)
        else:
            pfs = [prepare]
        for pf in pfs:
            config_dct = pf(config_dct)

    return msh.unmarshal_obj(config_dct, cls)



# @omlish-lite
# ruff: noqa: UP006 UP007 UP043 UP045
import dataclasses as dc
import typing as ta

from ...lite.abstract import Abstract


##


class CoroHttpIo:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    def __init_subclass__(cls, **kwargs):  # noqa
        raise TypeError

    #

    MAX_LINE: ta.ClassVar[int] = 65536

    #

    class Io(Abstract):
        pass

    #

    class AnyLogIo(Io, Abstract):
        pass

    #

    @dc.dataclass(frozen=True)
    class ConnectIo(Io):
        args: ta.Tuple[ta.Any, ...]
        kwargs: ta.Optional[ta.Dict[str, ta.Any]] = None

        server_hostname: ta.Optional[str] = None

    #

    class CloseIo(Io):
        pass

    #

    class AnyReadIo(Io):  # noqa
        pass

    @dc.dataclass(frozen=True)
    class ReadIo(AnyReadIo):
        sz: ta.Optional[int]

    @dc.dataclass(frozen=True)
    class ReadLineIo(AnyReadIo):
        sz: int

    @dc.dataclass(frozen=True)
    class PeekIo(AnyReadIo):
        sz: int

    #

    @dc.dataclass(frozen=True)
    class WriteIo(Io):
        data: bytes



import abc
import dataclasses as dc
import typing as ta

from .... import lang
from ..processing.base import ProcessingContext
from .ops import Op
from .ops import OpRefMap


T = ta.TypeVar('T')
PlanT = ta.TypeVar('PlanT')


##


@dc.dataclass(frozen=True)
class Plan(lang.Abstract):
    pass


##


@dc.dataclass(frozen=True)
class PlanResult(ta.Generic[PlanT]):
    plan: PlanT
    ref_map: OpRefMap | None = None


class Generator(lang.Abstract, ta.Generic[PlanT]):
    @abc.abstractmethod
    def plan(self, ctx: ProcessingContext) -> PlanResult[PlanT] | None:
        raise NotImplementedError

    @abc.abstractmethod
    def generate(self, pl: PlanT) -> ta.Iterable[Op]:
        raise NotImplementedError



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .clients.asyncs import (  # noqa
        AsyncStreamHttpResponse,

        async_close_http_client_response,
        async_closing_http_client_response,
        async_read_http_client_response,

        AsyncHttpClient,
    )

    from .clients.base import (  # noqa
        DEFAULT_ENCODING,
        is_success_status,

        HttpRequest,

        BaseHttpResponse,
        HttpResponse,

        HttpClientContext,

        HttpClientError,
        HttpStatusError,

        BaseHttpClient,
    )

    from .clients.default import (  # noqa
        client,
        manage_client,

        request,

        async_client,
        manage_async_client,

        async_request,
    )

    from .clients.httpx import (  # noqa
        HttpxHttpClient,

        HttpxAsyncHttpClient,
    )

    from .clients.middleware import (  # noqa
        HttpClientMiddleware,
        AbstractMiddlewareHttpClient,

        MiddlewareHttpClient,
        MiddlewareAsyncHttpClient,

        TooManyRedirectsHttpClientError,
        RedirectHandlingHttpClientMiddleware,
    )

    from .clients.sync import (  # noqa
        StreamHttpResponse,

        close_http_client_response,
        closing_http_client_response,
        read_http_client_response,

        HttpClient,
    )

    from .clients.syncasync import (  # noqa
        SyncAsyncHttpClient,
    )

    from .clients.urllib import (  # noqa
        UrllibHttpClient,
    )

    from . import consts  # noqa

    from .cookies import (  # noqa
        CookieTooBigError,
        dump_cookie,
        parse_cookie,
    )

    from .dates import (  # noqa
        http_date,
        parse_date,
    )

    from .encodings import (  # noqa
        latin1_decode,
        latin1_encode,
    )

    from .headers import (  # noqa
        CanHttpHeaders,
        DuplicateHttpHeaderError,
        HttpHeaders,
        headers,
    )

    from .json import (  # noqa
        JSON_TAGGER,
        JsonTag,
        JsonTagger,
        json_dumps,
        json_loads,
    )

    from .multipart import (  # noqa
        MultipartData,
        MultipartEncoder,
        MultipartField,
    )



# rufF: noqa: I001
from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .std.filters import (  # noqa
        TidLoggingFilter,
    )

    from .std.formatters import (  # noqa
        StdLoggingFormatter,
    )

    from .std.handlers import (  # noqa
        ListLoggingHandler,
    )

    from .std.json import (  # noqa
        JsonLoggingFormatter,
    )

    from .std.loggers import (  # noqa
        StdLogger,
    )

    from .std.noisy import (  # noqa
        silence_noisy_loggers,
    )

    from .std.proxy import (  # noqa
        ProxyLoggingFilterer,
        ProxyLoggingHandler,
    )

    from .std.records import (  # noqa
        LoggingContextLogRecord,
    )

    from .base import (  # noqa
        AnyLogger,
        Logger,
        AsyncLogger,

        AnyNopLogger,
        NopLogger,
        AsyncNopLogger,
    )

    from .contexts import (  # noqa
        LoggingContext,
    )

    from .formatters import (  # noqa
        LoggingContextFormatter,
    )

    from .infos import (  # noqa
        LoggingContextInfo,
        LoggingContextInfos,
    )

    from .levels import (  # noqa
        LogLevel,

        NamedLogLevel,
    )

    from .modules import (  # noqa
        get_module_logger,
    )

    from .protocols import (  # noqa
        LoggerLike,
    )

    from .standard import (  # noqa
        STANDARD_LOG_FORMAT_PARTS,
        StandardLoggingFormatter,

        StandardConfiguredLoggingHandler,

        configure_standard_logging,
    )

    from .utils import (  # noqa
        LogTimingContext,
        log_timing_context,

        error_logging,
    )

    from .warnings import (  # noqa
        LoggingSetupWarning,
    )



import dataclasses as dc
import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from .configs import Config


if ta.TYPE_CHECKING:
    from . import types as _types
    from .types import Marshaler
    from .types import MarshalerFactory
    from .types import Unmarshaler
    from .types import UnmarshalerFactory

else:
    _types = lang.proxy_import('.types', __package__)


##


@dc.dataclass(frozen=True, kw_only=True)
class Override(Config, lang.Final):
    marshaler: ta.Optional['Marshaler'] = None
    marshaler_factory: ta.Optional['MarshalerFactory'] = None

    unmarshaler: ta.Optional['Unmarshaler'] = None
    unmarshaler_factory: ta.Optional['UnmarshalerFactory'] = None

    def __post_init__(self) -> None:
        check.isinstance(self.marshaler, (_types.Marshaler, None))
        check.isinstance(self.unmarshaler, (_types.Unmarshaler, None))


@dc.dataclass(frozen=True)
class ReflectOverride(Config, lang.Final):
    rty: rfl.Type



import inspect
import typing as ta

from ... import check
from ... import collections as col
from ... import lang
from ... import reflect as rfl
from ..base.contexts import MarshalFactoryContext
from ..base.contexts import UnmarshalFactoryContext
from ..base.options import Option
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from .marshal import ObjectMarshaler
from .metadata import FieldInfo
from .metadata import FieldInfos
from .metadata import FieldMetadata
from .unmarshal import ObjectUnmarshaler


##


def _is_namedtuple(rty: rfl.Type) -> bool:
    return (
        isinstance(rty, type) and
        issubclass(rty, tuple) and
        ta.NamedTuple in rfl.get_orig_bases(rty)
    )


def get_namedtuple_field_infos(
        ty: type,
        opts: col.TypeMap[Option] = col.TypeMap(),
) -> FieldInfos:
    check.arg(_is_namedtuple(ty), ty)

    sig = inspect.signature(ty)

    ret: list[FieldInfo] = []
    for param in sig.parameters.values():
        ret.append(FieldInfo(
            name=param.name,
            type=param.annotation,
            metadata=FieldMetadata(),

            marshal_name=param.name,
            unmarshal_names=[param.name],
        ))

    return FieldInfos(ret)


##


class NamedtupleMarshalerFactory(MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not _is_namedtuple(rty):
            return None

        def inner() -> Marshaler:
            check.state(_is_namedtuple(rty))
            ty = check.isinstance(rty, type)
            check.state(not lang.is_abstract_class(ty))

            fis = get_namedtuple_field_infos(ty, ctx.options)

            fields = [
                (fi, ctx.make_marshaler(fi.type))
                for fi in fis
            ]

            return ObjectMarshaler(
                fields,
            )

        return inner


##


class NamedtupleUnmarshalerFactory(UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not _is_namedtuple(rty):
            return None

        def inner() -> Unmarshaler:
            check.state(_is_namedtuple(rty))
            ty = check.isinstance(rty, type)
            check.state(not lang.is_abstract_class(ty))

            fis = get_namedtuple_field_infos(ty, ctx.options)

            d: dict[str, tuple[FieldInfo, Unmarshaler]] = {}
            defaults: dict[str, ta.Any] = {}

            for fi in fis:
                tup = (fi, ctx.make_unmarshaler(fi.type))

                for un in fi.unmarshal_names:
                    if un in d:
                        raise KeyError(f'Duplicate fields for name {un!r}: {fi.name!r}, {d[un][0].name!r}')
                    d[un] = tup

                if fi.options.default.present:
                    defaults[fi.name] = fi.options.default.must()

            return ObjectUnmarshaler(
                ty,
                d,
                defaults=defaults,
            )

        return inner



import importlib
import string
import typing as ta


##


class ResolvableClassNameError(NameError):
    pass


def get_cls_fqcn(cls: type, *, nocheck: bool = False) -> str:
    if not isinstance(cls, type):
        raise TypeError(cls)

    mn = cls.__module__
    if set(mn) - set(string.ascii_lowercase + string.digits + '_.'):
        raise ResolvableClassNameError(cls)

    qn = cls.__qualname__
    if not all(qp[0].isupper() for qp in qn.split('.')) or (set(qn) - set(string.ascii_letters + string.digits + '.')):
        raise ResolvableClassNameError(cls)

    fqcn = '.'.join([cls.__module__, cls.__qualname__])
    if not nocheck:
        if get_fqcn_cls(fqcn, nocheck=True) is not cls:
            raise ResolvableClassNameError(cls, fqcn)

    return fqcn


def get_fqcn_cls(fqcn: str, *, nocheck: bool = False) -> type:
    if not isinstance(fqcn, str) or not fqcn:
        raise TypeError(fqcn)

    parts = fqcn.split('.')
    pos = next(i for i, p in enumerate(parts) if p[0].isupper())
    mps, qps = parts[:pos], parts[pos:]
    mod = importlib.import_module('.'.join(mps))

    o: ta.Any = mod
    for qp in qps:
        o = getattr(o, qp)
        if not isinstance(o, type):
            raise TypeError(o)

    cls = o
    if not isinstance(cls, type):
        raise TypeError(cls)

    if not nocheck:
        if not get_cls_fqcn(cls, nocheck=True) == fqcn:
            raise ResolvableClassNameError(cls, fqcn)

    return o


class Resolvable:
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        get_cls_fqcn(cls, nocheck=True)



import abc
import functools
import typing as ta

from ..attrstorage import transient_getattr
from ..attrstorage import transient_setattr
from ..classes.abstract import Abstract


_IGNORE = object()


##


class _CachedProperty(property, Abstract):
    def __init__(
            self,
            fn,
            *,
            name=None,
            ignore_if=lambda _: False,
    ):
        if isinstance(fn, property):
            fn = fn.fget

        super().__init__(fn)

        self._fn = fn
        self._ignore_if = ignore_if

        self._name = name

    def __set_name__(self, owner, name):
        if self._name is None:
            self._name = name

    @abc.abstractmethod
    def _instance_get(self, instance: ta.Any) -> tuple[ta.Any, bool]:
        raise NotImplementedError

    @abc.abstractmethod
    def _instance_set(self, instance: ta.Any, value: ta.Any) -> None:
        raise NotImplementedError

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self._name is None:
            raise TypeError(self)

        value, ok = self._instance_get(instance)
        if ok:
            return value

        value = self._fn.__get__(instance, owner)()
        if value is _IGNORE:
            return None

        self._instance_set(instance, value)
        return value

    def __set__(self, instance, value):
        if self._ignore_if(value):
            return

        ev, ok = self._instance_get(instance)
        if ok and ev == value:
            return

        raise TypeError(self._name)

    def __delete__(self, instance):
        raise TypeError


#


class _DictCachedProperty(_CachedProperty):
    def _instance_get(self, instance: ta.Any) -> tuple[ta.Any, bool]:
        try:
            value = instance.__dict__[self._name]
        except KeyError:
            return None, False
        else:
            return value, True

    def _instance_set(self, instance: ta.Any, value: ta.Any) -> None:
        instance.__dict__[self._name] = value


#


class _TransientCachedProperty(_CachedProperty):
    def _instance_get(self, instance: ta.Any) -> tuple[ta.Any, bool]:
        try:
            value = transient_getattr(instance, self._name)
        except AttributeError:
            return None, False
        else:
            return value, True

    def _instance_set(self, instance: ta.Any, value: ta.Any) -> None:
        transient_setattr(instance, self._name, value)


#


def cached_property(fn=None, *, transient=False, **kwargs):  # noqa
    if fn is None:
        return functools.partial(cached_property, transient=transient, **kwargs)

    elif transient:
        return _TransientCachedProperty(fn, **kwargs)

    else:
        return _DictCachedProperty(fn, **kwargs)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta

from ..io.buffers import ReadableListBuffer


##


class AsyncBufferedReader:
    def __init__(
            self,
            read: ta.Callable[[], ta.Awaitable[bytes]],
    ) -> None:
        super().__init__()

        self._read = read

        self._buf = ReadableListBuffer()

    class Cancelled(Exception):  # noqa
        pass

    async def read_until(
            self,
            delim: bytes,
            *,
            canceller: ta.Optional[ta.Callable[[], bool]] = None,
    ) -> bytes:
        o = 0
        while True:
            if canceller is not None and canceller():
                raise self.Cancelled
            r = self._buf.read_until_(delim, o)
            if isinstance(r, bytes):
                return r
            o = r
            b = await self._read()
            if not b:
                raise EOFError
            self._buf.feed(b)

    async def read_exact(
            self,
            sz: int,
            *,
            canceller: ta.Optional[ta.Callable[[], bool]] = None,
    ) -> bytes:
        while True:
            if canceller is not None and canceller():
                raise self.Cancelled
            r = self._buf.read(sz)
            if r is not None:
                return r
            b = await self._read()
            if not b:
                raise EOFError
            self._buf.feed(b)



import dataclasses as dc
import sys
import typing as ta

from ... import lang
from .base import Compression
from .codecs import make_compression_codec
from .codecs import make_compression_lazy_loaded_codec


if sys.version_info >= (3, 14):
    if ta.TYPE_CHECKING:
        from compression import zstd  # noqa
    else:
        zstd = lang.proxy_import('compression.zstd')

else:  # noqa
    if ta.TYPE_CHECKING:
        import zstandard as zstd
    else:
        zstd = lang.proxy_import('zstandard')


##


@dc.dataclass(frozen=True, kw_only=True)
class ZstdCompression(Compression):
    level: int | None = None

    # max_output_size: int = 0

    def compress(self, d: bytes) -> bytes:
        return zstd.compress(
            d,
            **(dict(level=self.level) if self.level is not None else {}),
        )

    def decompress(self, d: bytes) -> bytes:
        return zstd.decompress(
            d,
            # max_output_size=self.max_output_size,
        )


##


ZSTD_CODEC = make_compression_codec('zstd', ZstdCompression)

# @omlish-manifest
_ZSTD_LAZY_CODEC = make_compression_lazy_loaded_codec(__name__, 'ZSTD_CODEC', ZSTD_CODEC)



import typing as ta


O = ta.TypeVar('O')
I = ta.TypeVar('I')
R = ta.TypeVar('R')


# Direct coros yield outputs 1:1 with inputs.
DirectCoro: ta.TypeAlias = ta.Generator[O, I, R]

BytesDirectCoro: ta.TypeAlias = DirectCoro[bytes, bytes, R]
StrDirectCoro: ta.TypeAlias = DirectCoro[str, str, R]



# ruff: noqa: UP007
# @omlish-lite
import logging
import typing as ta

from ..base import Logger
from ..base import LoggingMsgFn
from ..contexts import CaptureLoggingContext
from ..infos import LoggingContextInfos
from ..levels import LogLevel
from .records import LoggingContextLogRecord


##


class StdLogger(Logger):
    def __init__(self, std: logging.Logger) -> None:
        super().__init__()

        self._std = std

    @property
    def std(self) -> logging.Logger:
        return self._std

    def is_enabled_for(self, level: LogLevel) -> bool:
        return self._std.isEnabledFor(level)

    def get_effective_level(self) -> LogLevel:
        return self._std.getEffectiveLevel()

    def _log(self, ctx: CaptureLoggingContext, msg: ta.Union[str, tuple, LoggingMsgFn], *args: ta.Any) -> None:
        if not self.is_enabled_for(ctx.must_get_info(LoggingContextInfos.Level).level):
            return

        ctx.set_basic(
            name=self._std.name,

            msg=msg,
            args=args,
        )

        ctx.capture()

        rec = LoggingContextLogRecord(_logging_context=ctx)

        self._std.handle(rec)



from .imports.proxy import auto_proxy_init as _auto_proxy_init


with _auto_proxy_init(globals(), update_exports=True):
    ##

    from .asyncs import (  # noqa
        as_async,

        AsyncGeneratorWithReturn,
        async_generator_with_return,
    )

    from .attrstorage import (  # noqa
        AttributePresentError,
        SetAttrIfPresent,
        set_attr,

        AttrStorage,

        StdAttrStorage,
        STD_ATTR_STORAGE,

        DictAttrStorage,

        TransientDict,
        TransientAttrStorage,
        TRANSIENT_ATTR_STORAGE,
        transient_delattr,
        transient_getattr,
        transient_setattr,
    )

    from .cached.function import (  # noqa
        cached_function,
        static_init,
    )

    from .cached.property import (  # noqa
        cached_property,
    )

    from .casing import (  # noqa
        StringCasingError,
        ImproperStringCasingError,
        UnknownStringCasingError,
        AmbiguousStringCasingError,

        StringCasing,
        CamelCase,
        LowCamelCase,
        SnakeCase,
        UpSnakeCase,

        STRING_CASINGS,
        CAMEL_CASE,
        LOW_CAMEL_CASE,
        SNAKE_CASE,
        UP_SNAKE_CASE,

        camel_case,
        low_camel_case,
        snake_case,
        up_snake_case,

        camel_to_snake,
        snake_to_camel,

        get_string_casing,
        split_string_casing,
    )

    from .classes.abstract import (  # noqa
        is_abstract_method,

        is_abstract_class,
        is_abstract,

        get_abstracts,
        make_abstract,
    )

    from .classes.bindable import (  # noqa
        BindableClass,
    )

    from .classes.namespaces import (  # noqa
        GenericNamespaceMeta,
        NamespaceMeta,
        Namespace,
    )

    from .classes.protocols import (  # noqa
        ProtocolForbiddenAsBaseClass,
        ProtocolForbiddenAsBaseClassTypeError,
    )

    from .classes.restrict import (  # noqa
        FinalTypeError,
        Final,

        SealedError,
        Sealed,
        PackageSealed,

        NotInstantiable,
        NotPicklable,

        NoBool,
        no_bool,

        SENSITIVE_ATTR,
        Sensitive,
        AnySensitive,
    )

    from .classes.simple import (  # noqa
        SimpleMetaDict,

        Marker,

        Singleton,
        LazySingleton,
    )

    from .classes.virtual import (  # noqa
        Virtual,
        virtual_check,

        Descriptor,
        Picklable,

        Callable,
    )

    from .clsdct import (  # noqa
        is_possibly_cls_dct,
        get_caller_cls_dct,

        ClassDctFn,
        cls_dct_fn,
    )

    from .collections import (  # noqa
        yield_dict_init,
        merge_dicts,

        empty_map,
    )

    from .comparison import (  # noqa
        cmp,

        InfinityType,
        Infinity,
        NegativeInfinityType,
        NegativeInfinity,
    )

    from .contextmanagers import (  # noqa
        ContextManaged,
        SelfContextManaged,
        ValueContextManager,
        NOP_CONTEXT_MANAGER,

        AsyncContextManaged,
        SelfAsyncContextManaged,
        ValueAsyncContextManager,
        NOP_ASYNC_CONTEXT_MANAGER,

        AsyncContextManager,

        ContextManager,

        maybe_managing,
        async_maybe_managing,
        async_or_sync_maybe_managing,

        disposing,
        breakpoint_on_exception,
        context_var_setting,

        ContextWrappable,
        ContextWrapped,
        context_wrapped,

        Lockable,
        DefaultLockable,
        default_lock,

        AsyncLockable,
        DefaultAsyncLockable,
        default_async_lock,

        Timer,

        double_check_setdefault,

        call_with_exit_stack,
        call_with_async_exit_stack,

        with_exit_stack,
        with_async_exit_stack,
    )

    from .datetimes import (  # noqa
        ISO_FMT,
        ISO_FMT_US,
        ISO_FMT_NTZ,
        ISO_FMT_US_NTZ,

        utcnow,
        utcfromtimestamp,
    )

    from .descriptors import (  # noqa
        attr_property,
        item_property,

        is_method_descriptor,
        unwrap_method_descriptors,

        unwrap_func_with_partials,
        unwrap_func,

        unwrap_callable_with_partials,
        unwrap_callable,

        update_wrapper,

        decorator,

        AccessForbiddenError,
        access_forbidden,

        classonly,
    )

    from .enums import (  # noqa
        enum_name_repr,
    )

    from .errors import (  # noqa
        DuplicateKeyError,
    )

    from .functions import (  # noqa
        is_lambda,

        call_with,
        opt_call,
        opt_fn,
        recurse,

        raise_,
        raising,
        try_,
        finally_,

        identity,
        constant,
        nullary_constant,

        is_none,
        is_not_none,
        isinstance_of,
        issubclass_of,
        strict_eq,

        VoidError,
        Void,
        void,

        periodically,

        opt_getattr,
        coalesce,
        opt_coalesce,

        cond_kw,
        opt_kw,
        truthy_kw,

        new_function,
        new_function_kwargs,
        copy_function,
    )

    from .generators import (  # noqa
        nextgen,
        autostart,

        GeneratorLike,

        capture_generator,
        capture_coroutine,

        GeneratorMappedIterator,
        genmap,
    )

    from .imports.capture import (  # noqa
        ImportCaptureError,
        ImportCaptureErrors,
        ImportCapture,
    )

    from .imports.conditional import (  # noqa
        register_conditional_import,
        trigger_conditional_imports,
    )

    from .imports.lazy import (  # noqa
        lazy_import,
    )

    from .imports.proxy import (  # noqa
        proxy_import,
        auto_proxy_import,

        proxy_init,
        auto_proxy_init,
    )

    from .imports.resolving import (  # noqa
        can_import,
        get_real_module_name,
        resolve_import_name,
        try_import,
    )

    from .imports.traversal import (  # noqa
        import_all,
        yield_import_all,
        yield_importable,
    )

    from .iterables import (  # noqa
        ilen,
        take,
        consume,
        peek,
        chunk,
        interleave,
        renumerate,
        common_prefix_len,

        readiter,

        itergen,

        flatten,
        flatmap,

        asrange,
        prodrange,

        IteratorWithReturn,
    )

    from .lazyglobals import (  # noqa
        AmbiguousLazyGlobalsFallbackError,
        LazyGlobals,
    )

    from .maybes import (  # noqa
        empty,
        just,
    )

    from .maysync import (  # noqa
        make_maysync_fn,
        make_maysync_generator_fn,
        make_maysync,

        make_maysync_from_sync,
    )

    from .objects import (  # noqa
        arg_repr,
        opt_repr,

        can_weakref,

        new_type,
        super_meta,

        deep_subclasses,

        SimpleProxy,

        anon_object,

        Identity,
    )

    from .outcomes import (  # noqa
        OutcomeAlreadyUnwrappedError,

        capture,
        acapture,

        Outcome,

        Value,
        value,

        Error,
        error,

        Either,
    )

    from .overrides import (  # noqa
        needs_override,

        is_override,

        RequiresOverrideError,
        RequiresOverride,
    )

    from .params import (  # noqa
        CanParamSpec,

        Param,

        VarParam,
        ArgsParam,
        KwargsParam,

        ValParam,
        PosOnlyParam,
        KwOnlyParam,

        ParamSeparator,

        ParamSpec,

        param_render,
    )

    from .recursion import (  # noqa
        LimitedRecursionError,
        recursion_limiting_context,

        recursion_limiting,
    )

    from .resolving import (  # noqa
        ResolvableClassNameError,
        get_cls_fqcn,
        get_fqcn_cls,
        Resolvable,
    )

    from .resources import (  # noqa
        ReadableResource,
        get_package_resources,
        get_relative_resources,
    )

    from .sequences import (  # noqa
        iterslice,
        iterrange,

        SeqView,
    )

    from .strings import (  # noqa
        prefix_delimited,
        prefix_lines,
        indent_lines,

        strip_prefix,
        strip_suffix,

        replace_many,

        find_any,
        rfind_any,

        is_dunder,
        is_sunder,

        is_ident_start,
        is_ident_cont,
        is_ident,

        BOOL_STRINGS,
        BOOL_FALSE_STRINGS,
        BOOL_TRUE_STRINGS,
        STRING_BOOL_VALUES,

        iter_pat,
    )

    from .sys import (  # noqa
        REQUIRED_PYTHON_VERSION,
        check_runtime_version,
        is_gil_enabled,
    )

    from .typing import (  # noqa
        BytesLike,

        static_check_isinstance,
        static_check_issubclass,
        copy_type,

        protocol_check,

        typed_lambda,
        typed_partial,

        SequenceNotStr,
    )

    ##

    from ..lite.abstract import (  # noqa
        is_abstract_method,
        update_abstracts,

        AbstractTypeError,
        Abstract,
    )

    from ..lite.args import (  # noqa
        Args,
    )

    from ..lite.asyncs import (  # noqa
        opt_await,
        async_list,
        async_enumerate,

        SyncAwaitCoroutineNotTerminatedError,
        sync_await,

        sync_aiter,
        sync_async_list,

        SyncAwaitContextManager,
        sync_async_with,

        SyncToAsyncContextManager,
        as_async_context_manager,
    )

    from ..lite.attrops import (  # noqa
        AttrOps,
        attr_ops,

        attr_repr,
    )

    from ..lite.contextmanagers import (  # noqa
        AsyncExitStacked,
        ExitStacked,
        adefer,
        attr_setting,
        defer,
    )

    from ..lite.imports import (  # noqa
        import_attr,
        import_module,
        import_module_attr,
    )

    from ..lite.maybes import (  # noqa
        Maybe,
    )

    from ..lite.maysync import (  # noqa
        AnyMaysyncFn,

        MaywaitableAlreadyConsumedError,
        AnyMaywaitable,

        MaysyncFn,
        Maywaitable,

        MaysyncGeneratorFn,
        MaysyncGenerator,

        is_running_maysync,

        run_maysync,

        RunMaysyncContextManager,
        run_maysync_context_manager,

        mark_maysync,
        is_maysync,
    )

    from ..lite.objects import (  # noqa
        mro_dict,
        mro_owner_dict,
        dir_dict,
    )

    from ..lite.timeouts import (  # noqa
        DeadlineTimeout,
        InfiniteTimeout,
        Timeout,
        TimeoutLike,
    )

    from ..lite.types import (  # noqa
        BUILTIN_SCALAR_ITERABLE_TYPES,
    )

    from ..lite.typing import (  # noqa
        AnyFunc,
        Func0,
        Func1,
        Func2,
        Func3,

        typing_annotations_attr,
    )

    from ..lite.wrappers import (  # noqa
        update_wrapper_no_annotations,
    )



import typing as ta

from .. import dataclasses as dc
from .keys import Key
from .types import Scope


##


@dc.dataclass()
class BaseKeyError(Exception):
    key: Key

    source: ta.Any = None
    name: str | None = None


@dc.dataclass()
class UnboundKeyError(BaseKeyError):
    pass


@dc.dataclass()
class ConflictingKeyError(BaseKeyError):
    pass


@dc.dataclass()
class CyclicDependencyError(BaseKeyError):
    pass


##


@dc.dataclass()
class ScopeError(Exception):
    scope: Scope


@dc.dataclass()
class ScopeAlreadyOpenError(ScopeError):
    pass


@dc.dataclass()
class ScopeNotOpenError(ScopeError):
    pass



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - remove redundant info fields only present for std adaptation (Level.name, ...)
"""
import collections.abc
import io
import logging
import os.path
import sys
import threading
import time
import traceback
import types
import typing as ta

from .levels import NamedLogLevel
from .warnings import LoggingSetupWarning


LoggingMsgFn = ta.Callable[[], ta.Union[str, tuple]]  # ta.TypeAlias

LoggingExcInfoTuple = ta.Tuple[ta.Type[BaseException], BaseException, ta.Optional[types.TracebackType]]  # ta.TypeAlias
LoggingExcInfo = ta.Union[BaseException, LoggingExcInfoTuple]  # ta.TypeAlias
LoggingExcInfoArg = ta.Union[LoggingExcInfo, bool, None]  # ta.TypeAlias

LoggingContextInfo = ta.Any  # ta.TypeAlias


##


def logging_context_info(cls):
    return cls


@ta.final
class LoggingContextInfos:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    #

    @logging_context_info
    @ta.final
    class Name(ta.NamedTuple):
        name: str

    @logging_context_info
    @ta.final
    class Level(ta.NamedTuple):
        level: NamedLogLevel
        name: str

        @classmethod
        def build(cls, level: int) -> 'LoggingContextInfos.Level':
            nl: NamedLogLevel = level if level.__class__ is NamedLogLevel else NamedLogLevel(level)  # type: ignore[assignment]  # noqa
            return cls(
                level=nl,
                name=logging.getLevelName(nl),
            )

    @logging_context_info
    @ta.final
    class Msg(ta.NamedTuple):
        msg: str
        args: ta.Union[tuple, ta.Mapping[ta.Any, ta.Any], None]

        @classmethod
        def build(
                cls,
                msg: ta.Union[str, tuple, LoggingMsgFn],
                *args: ta.Any,
        ) -> 'LoggingContextInfos.Msg':
            s: str
            a: ta.Any

            if callable(msg):
                if args:
                    raise TypeError(f'Must not provide both a message function and args: {msg=} {args=}')
                x = msg()
                if isinstance(x, str):
                    s, a = x, ()
                elif isinstance(x, tuple):
                    if x:
                        s, a = x[0], x[1:]
                    else:
                        s, a = '', ()
                else:
                    raise TypeError(x)

            elif isinstance(msg, tuple):
                if args:
                    raise TypeError(f'Must not provide both a tuple message and args: {msg=} {args=}')
                if msg:
                    s, a = msg[0], msg[1:]
                else:
                    s, a = '', ()

            elif isinstance(msg, str):
                s, a = msg, args

            else:
                raise TypeError(msg)

            # https://github.com/python/cpython/blob/e709361fc87d0d9ab9c58033a0a7f2fef0ad43d2/Lib/logging/__init__.py#L307  # noqa
            if a and len(a) == 1 and isinstance(a[0], collections.abc.Mapping) and a[0]:
                a = a[0]

            return cls(
                msg=s,
                args=a,
            )

    @logging_context_info
    @ta.final
    class Extra(ta.NamedTuple):
        extra: ta.Mapping[ta.Any, ta.Any]

    @logging_context_info
    @ta.final
    class Time(ta.NamedTuple):
        ns: int
        secs: float
        msecs: float
        relative_secs: float

        @classmethod
        def get_std_start_ns(cls) -> int:
            x: ta.Any = logging._startTime  # type: ignore[attr-defined]  # noqa

            # Before 3.13.0b1 this will be `time.time()`, a float of seconds. After that, it will be `time.time_ns()`,
            # an int.
            #
            # See:
            #  - https://github.com/python/cpython/commit/1316692e8c7c1e1f3b6639e51804f9db5ed892ea
            #
            if isinstance(x, float):
                return int(x * 1e9)
            else:
                return x

        @classmethod
        def build(
                cls,
                ns: int,
                *,
                start_ns: ta.Optional[int] = None,
        ) -> 'LoggingContextInfos.Time':
            # https://github.com/python/cpython/commit/1316692e8c7c1e1f3b6639e51804f9db5ed892ea
            secs = ns / 1e9  # ns to float seconds

            # Get the number of whole milliseconds (0-999) in the fractional part of seconds.
            # Eg: 1_677_903_920_999_998_503 ns --> 999_998_503 ns--> 999 ms
            # Convert to float by adding 0.0 for historical reasons. See gh-89047
            msecs = (ns % 1_000_000_000) // 1_000_000 + 0.0

            # https://github.com/python/cpython/commit/1500a23f33f5a6d052ff1ef6383d9839928b8ff1
            if msecs == 999.0 and int(secs) != ns // 1_000_000_000:
                # ns -> sec conversion can round up, e.g:
                # 1_677_903_920_999_999_900 ns --> 1_677_903_921.0 sec
                msecs = 0.0

            if start_ns is None:
                start_ns = cls.get_std_start_ns()
            relative_secs = (ns - start_ns) / 1e6

            return cls(
                ns=ns,
                secs=secs,
                msecs=msecs,
                relative_secs=relative_secs,
            )

    @logging_context_info
    @ta.final
    class Exc(ta.NamedTuple):
        info: LoggingExcInfo
        info_tuple: LoggingExcInfoTuple

        @classmethod
        def build(
                cls,
                arg: LoggingExcInfoArg = False,
        ) -> ta.Optional['LoggingContextInfos.Exc']:
            if arg is True:
                sys_exc_info = sys.exc_info()
                if sys_exc_info[0] is not None:
                    arg = sys_exc_info
                else:
                    arg = None
            elif arg is False:
                arg = None
            if arg is None:
                return None

            info: LoggingExcInfo = arg
            if isinstance(info, BaseException):
                info_tuple: LoggingExcInfoTuple = (type(info), info, info.__traceback__)  # noqa
            else:
                info_tuple = info

            return cls(
                info=info,
                info_tuple=info_tuple,
            )

    @logging_context_info
    @ta.final
    class Caller(ta.NamedTuple):
        file_path: str
        line_no: int
        func_name: str
        stack_info: ta.Optional[str]

        @classmethod
        def is_internal_frame(cls, frame: types.FrameType) -> bool:
            file_path = os.path.normcase(frame.f_code.co_filename)

            # Yes, really.
            # https://github.com/python/cpython/blob/e709361fc87d0d9ab9c58033a0a7f2fef0ad43d2/Lib/logging/__init__.py#L204  # noqa
            # https://github.com/python/cpython/commit/5ca6d7469be53960843df39bb900e9c3359f127f
            if 'importlib' in file_path and '_bootstrap' in file_path:
                return True

            return False

        @classmethod
        def find_frame(cls, stack_offset: int = 0) -> ta.Optional[types.FrameType]:
            f: ta.Optional[types.FrameType] = sys._getframe(2 + stack_offset)  # noqa

            while f is not None:
                # NOTE: We don't check __file__ like stdlib since we may be running amalgamated - we rely on careful,
                # manual stack_offset management.
                if hasattr(f, 'f_code'):
                    return f

                f = f.f_back

            return None

        @classmethod
        def build(
                cls,
                stack_offset: int = 0,
                *,
                stack_info: bool = False,
        ) -> ta.Optional['LoggingContextInfos.Caller']:
            if (f := cls.find_frame(stack_offset + 1)) is None:
                return None

            # https://github.com/python/cpython/blob/08e9794517063c8cd92c48714071b1d3c60b71bd/Lib/logging/__init__.py#L1616-L1623  # noqa
            sinfo = None
            if stack_info:
                sio = io.StringIO()
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                sio.close()
                if sinfo[-1] == '\n':
                    sinfo = sinfo[:-1]

            return cls(
                file_path=f.f_code.co_filename,
                line_no=f.f_lineno or 0,
                func_name=f.f_code.co_name,
                stack_info=sinfo,
            )

    @logging_context_info
    @ta.final
    class SourceFile(ta.NamedTuple):
        file_name: str
        module: str

        @classmethod
        def build(cls, caller_file_path: ta.Optional[str]) -> ta.Optional['LoggingContextInfos.SourceFile']:
            if caller_file_path is None:
                return None

            # https://github.com/python/cpython/blob/e709361fc87d0d9ab9c58033a0a7f2fef0ad43d2/Lib/logging/__init__.py#L331-L336  # noqa
            try:
                file_name = os.path.basename(caller_file_path)
                module = os.path.splitext(file_name)[0]
            except (TypeError, ValueError, AttributeError):
                return None

            return cls(
                file_name=file_name,
                module=module,
            )

    @logging_context_info
    @ta.final
    class Thread(ta.NamedTuple):
        ident: int
        native_id: ta.Optional[int]
        name: str

        @classmethod
        def build(cls) -> 'LoggingContextInfos.Thread':
            return cls(
                ident=threading.get_ident(),
                native_id=threading.get_native_id() if hasattr(threading, 'get_native_id') else None,
                name=threading.current_thread().name,
            )

    @logging_context_info
    @ta.final
    class Process(ta.NamedTuple):
        pid: int

        @classmethod
        def build(cls) -> 'LoggingContextInfos.Process':
            return cls(
                pid=os.getpid(),
            )

    @logging_context_info
    @ta.final
    class Multiprocessing(ta.NamedTuple):
        process_name: str

        @classmethod
        def build(cls) -> ta.Optional['LoggingContextInfos.Multiprocessing']:
            # https://github.com/python/cpython/blob/e709361fc87d0d9ab9c58033a0a7f2fef0ad43d2/Lib/logging/__init__.py#L355-L364  # noqa
            if (mp := sys.modules.get('multiprocessing')) is None:
                return None

            return cls(
                process_name=mp.current_process().name,
            )

    @logging_context_info
    @ta.final
    class AsyncioTask(ta.NamedTuple):
        name: str

        @classmethod
        def build(cls) -> ta.Optional['LoggingContextInfos.AsyncioTask']:
            # https://github.com/python/cpython/blob/e709361fc87d0d9ab9c58033a0a7f2fef0ad43d2/Lib/logging/__init__.py#L372-L377  # noqa
            if (asyncio := sys.modules.get('asyncio')) is None:
                return None

            try:
                task = asyncio.current_task()
            except Exception:  # noqa
                return None

            if task is None:
                return None

            return cls(
                name=task.get_name(),  # Always non-None
            )


##


class UnexpectedLoggingStartTimeWarning(LoggingSetupWarning):
    pass


def _check_logging_start_time() -> None:
    if (x := LoggingContextInfos.Time.get_std_start_ns()) < (t := time.time()):
        import warnings  # noqa

        warnings.warn(
            f'Unexpected logging start time detected: '
            f'get_std_start_ns={x}, '
            f'time.time()={t}',
            UnexpectedLoggingStartTimeWarning,
        )


_check_logging_start_time()



import dataclasses as dc
import types
import typing as ta

from ..... import check
from ...._internals import STD_FIELDS_ATTR
from ...._internals import StdFieldType
from ...._internals import std_is_classvar
from ...._internals import std_is_initvar
from ...._internals import std_is_kw_only
from ....inspect import get_cls_annotations


##


def build_std_field(
        cls: type,
        a_name: str,
        a_type: ta.Any,
        *,
        default_kw_only: bool,
) -> dc.Field:
    default: ta.Any = getattr(cls, a_name, dc.MISSING)
    if isinstance(default, dc.Field):
        f = default
    else:
        if isinstance(default, types.MemberDescriptorType):
            # This is a field in __slots__, so it has no default value.
            default = dc.MISSING
        f = dc.field(default=default)

    f.name = a_name
    f.type = a_type

    # field type

    ft = StdFieldType.INSTANCE
    if std_is_classvar(cls, f.type):
        ft = StdFieldType.CLASS_VAR
    if std_is_initvar(cls, f.type):
        ft = StdFieldType.INIT_VAR
    if ft in (StdFieldType.CLASS_VAR, StdFieldType.INIT_VAR):
        if f.default_factory is not dc.MISSING:
            raise TypeError(f'field {a_name} cannot have a default factory')
    f._field_type = ft.value  # type: ignore[attr-defined]  # noqa

    # kw_only

    if ft in (StdFieldType.INSTANCE, StdFieldType.INIT_VAR):
        if f.kw_only is dc.MISSING:
            f.kw_only = default_kw_only
    else:
        check.arg(ft is StdFieldType.CLASS_VAR)
        if f.kw_only is not dc.MISSING:
            raise TypeError(f'field {a_name} is a ClassVar but specifies kw_only')

    # defaults

    if (
            ft is StdFieldType.INSTANCE and
            f.default is not dc.MISSING and
            f.default.__class__.__hash__ is None
    ):
        raise ValueError(f'mutable default {type(f.default)} for field {a_name} is not allowed: use default_factory')

    #

    return f


##


def build_cls_std_fields(
        cls: type,
        *,
        kw_only: bool,
) -> ta.Mapping[str, dc.Field]:
    fields: dict[str, dc.Field] = {}

    for b in cls.__mro__[-1:0:-1]:
        if not (base_fields := getattr(b, STD_FIELDS_ATTR, None)):
            continue
        for f in base_fields.values():
            fields[f.name] = f

    cls_annotations = get_cls_annotations(cls)

    cls_fields: list[dc.Field] = []

    kw_only_seen = False
    for name, ann in cls_annotations.items():
        if std_is_kw_only(cls, ann):
            if kw_only_seen:
                raise TypeError(f'{name!r} is KW_ONLY, but KW_ONLY has already been specified')
            kw_only_seen = True
            kw_only = True

        else:
            cls_fields.append(build_std_field(
                cls,
                name,
                ann,
                default_kw_only=kw_only,
            ))

    for f in cls_fields:
        fields[f.name] = f
        if isinstance(getattr(cls, f.name, None), dc.Field):
            if f.default is dc.MISSING:
                delattr(cls, f.name)
            else:
                setattr(cls, f.name, f.default)

    for name, value in cls.__dict__.items():
        if isinstance(value, dc.Field) and name not in cls_annotations:
            raise TypeError(f'{name!r} is a field but has no type annotation')

    return fields



# ruff: noqa: UP043 UP045
# @omlish-lite
import abc
import contextlib
import dataclasses as dc
import io
import typing as ta

from ...lite.abstract import Abstract
from ...lite.dataclasses import dataclass_shallow_asdict
from .base import BaseHttpClient
from .base import BaseHttpResponse
from .base import BaseHttpResponseT
from .base import HttpClientContext
from .base import HttpRequest
from .base import HttpResponse
from .base import HttpStatusError


AsyncStreamHttpResponseT = ta.TypeVar('AsyncStreamHttpResponseT', bound='AsyncStreamHttpResponse')
AsyncHttpClientT = ta.TypeVar('AsyncHttpClientT', bound='AsyncHttpClient')


##


@ta.final
@dc.dataclass(frozen=True)  # kw_only=True
class AsyncStreamHttpResponse(BaseHttpResponse):
    class Stream(ta.Protocol):
        def read1(self, /, n: int = -1) -> ta.Awaitable[bytes]: ...

    @ta.final
    class _NullStream:
        def read1(self, /, n: int = -1) -> ta.Awaitable[bytes]:
            raise TypeError

    stream: Stream = _NullStream()

    @property
    def has_data(self) -> bool:
        return not isinstance(self.stream, AsyncStreamHttpResponse._NullStream)

    async def read_all(self) -> bytes:
        buf = io.BytesIO()
        while (b := await self.stream.read1()):
            buf.write(b)
        return buf.getvalue()

    _closer: ta.Optional[ta.Callable[[], ta.Awaitable[None]]] = None

    async def __aenter__(self: AsyncStreamHttpResponseT) -> AsyncStreamHttpResponseT:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self) -> None:
        if (c := self._closer) is not None:
            await c()  # noqa


#


async def async_close_http_client_response(resp: BaseHttpResponse) -> None:
    if isinstance(resp, HttpResponse):
        pass

    elif isinstance(resp, AsyncStreamHttpResponse):
        await resp.close()

    else:
        raise TypeError(resp)


@contextlib.asynccontextmanager
async def async_closing_http_client_response(resp: BaseHttpResponseT) -> ta.AsyncGenerator[BaseHttpResponseT, None]:
    if isinstance(resp, HttpResponse):
        yield resp
        return

    elif isinstance(resp, AsyncStreamHttpResponse):
        try:
            yield resp
        finally:
            await resp.close()

    else:
        raise TypeError(resp)


async def async_read_http_client_response(resp: BaseHttpResponse) -> HttpResponse:
    if isinstance(resp, HttpResponse):
        return resp

    elif isinstance(resp, AsyncStreamHttpResponse):
        return HttpResponse(**{
            **{k: v for k, v in dataclass_shallow_asdict(resp).items() if k not in ('stream', '_closer')},
            **({'data': await resp.read_all()} if resp.has_data else {}),
        })

    else:
        raise TypeError(resp)


##


class AsyncHttpClient(BaseHttpClient, Abstract):
    async def __aenter__(self: AsyncHttpClientT) -> AsyncHttpClientT:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    async def request(
            self,
            req: HttpRequest,
            *,
            context: ta.Optional[HttpClientContext] = None,
            check: bool = False,
    ) -> HttpResponse:
        async with async_closing_http_client_response(await self.stream_request(
                req,
                context=context,
                check=check,
        )) as resp:
            return await async_read_http_client_response(resp)

    async def stream_request(
            self,
            req: HttpRequest,
            *,
            context: ta.Optional[HttpClientContext] = None,
            check: bool = False,
    ) -> AsyncStreamHttpResponse:
        if context is None:
            context = HttpClientContext()

        resp = await self._stream_request(context, req)

        try:
            if check and not resp.is_success:
                if isinstance(resp.underlying, Exception):
                    cause = resp.underlying
                else:
                    cause = None
                raise HttpStatusError(await async_read_http_client_response(resp)) from cause  # noqa

        except Exception:
            await async_close_http_client_response(resp)
            raise

        return resp

    @abc.abstractmethod
    def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> ta.Awaitable[AsyncStreamHttpResponse]:
        raise NotImplementedError



"""
TODO:
 - Accessor inputs/outputs should *ideally* be subtype of class generic param
  - but can't make a typevar with a bound of another typevar
"""
import abc
import typing as ta

from .. import check
from .. import lang
from .values import TypedValue  # noqa
from .values import UniqueTypedValue  # noqa


T = ta.TypeVar('T')

TypedValueT0 = ta.TypeVar('TypedValueT0', bound='TypedValue')
TypedValueU = ta.TypeVar('TypedValueU', bound='TypedValue')

UniqueTypedValueU = ta.TypeVar('UniqueTypedValueU', bound='UniqueTypedValue')


##


class _NOT_SET(lang.Marker):  # noqa
    pass


class TypedValuesAccessor(
    lang.Abstract,
    ta.Sequence[TypedValueT0],
    ta.Generic[TypedValueT0],
):
    def __iter__(self):
        raise TypeError(
            'TypedValuesAccessor does not implement __iter__ - it is reserved for implementation by subclasses.',
        )

    #

    @ta.final
    def __contains__(self, cls: type[TypedValueU]) -> bool:  # type: ignore[override]
        return self._typed_value_contains(cls)

    @abc.abstractmethod
    def _typed_value_contains(self, cls):
        raise NotImplementedError

    #

    @ta.overload  # type: ignore[override]
    def __getitem__(self, idx: int) -> TypedValueT0:
        ...

    @ta.overload
    def __getitem__(self, cls: type[UniqueTypedValueU]) -> UniqueTypedValueU:  # type: ignore[overload-overlap]
        ...

    @ta.overload
    def __getitem__(self, cls: type[TypedValueU]) -> ta.Sequence[TypedValueU]:
        ...

    @ta.final
    def __getitem__(self, key):
        return self._typed_value_getitem(key)

    @abc.abstractmethod
    def _typed_value_getitem(self, key):
        raise NotImplementedError

    #

    @ta.overload
    def get(
            self,
            tv: UniqueTypedValueU,
    ) -> UniqueTypedValueU:
        ...

    @ta.overload
    def get(
            self,
            cls: type[UniqueTypedValueU],
            /,
            default: UniqueTypedValueU,
    ) -> UniqueTypedValueU:
        ...

    @ta.overload
    def get(  # type: ignore[overload-overlap]
            self,
            cls: type[UniqueTypedValueU],
            /,
            default: None = None,
    ) -> UniqueTypedValueU | None:
        ...

    @ta.overload
    def get(
            self,
            cls: type[TypedValueU],
            /,
            default: ta.Iterable[TypedValueU] | None = None,
    ) -> ta.Sequence[TypedValueU]:
        ...

    @ta.final
    def get(self, key, /, default=_NOT_SET):
        if not isinstance(key, type):
            if default is not _NOT_SET:
                raise RuntimeError('Must not provide both an instance key and a default')
            default = key
            key = type(default)
        elif default is _NOT_SET:
            default = None

        check.issubclass(key, TypedValue)
        try:
            return self._typed_value_getitem(key)
        except KeyError:
            return default

    #

    @ta.final
    def get_any(self, cls: type[T] | tuple[type[T], ...]) -> ta.Sequence[T]:
        return self._typed_value_get_any(cls)

    @abc.abstractmethod
    def _typed_value_get_any(self, cls):
        raise NotImplementedError



"""
TODO:
 - if_not_single? switches does this
"""
import shutil
import sys
import sysconfig
import typing as ta

import pytest

from ... import lang


##


def if_cant_import(*modules: str, **kwargs):
    missing = [m for m in modules if not lang.can_import(m, **kwargs)]
    return pytest.mark.skipif(
        bool(missing),
        reason=f'requires import {", ".join(missing)}',
    )


def if_not_on_path(exe: str):
    return pytest.mark.skipif(
        shutil.which(exe) is None,
        reason=f'requires exe on path {exe}',
    )


def if_python_version_less_than(num: ta.Sequence[int]):
    return pytest.mark.skipif(
        sys.version_info < tuple(num),
        reason=f'python version {tuple(sys.version_info)} < {tuple(num)}',
    )


def if_not_platform(*platforms: str):
    return pytest.mark.skipif(
        sys.platform not in platforms,
        reason=f'requires platform in {platforms}',
    )


def if_nogil():
    return pytest.mark.skipif(
        sysconfig.get_config_var('Py_GIL_DISABLED'),
        reason='requires gil build',
    )



import typing as ta

from ... import lang
from ... import reflect as rfl
from ...funcs import guard as gfs
from ..base.contexts import MarshalFactoryContext
from ..base.contexts import UnmarshalFactoryContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory


##


class MarshalerFactoryMethodClass(MarshalerFactory, lang.Abstract):
    @gfs.method(instance_cache=True)
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        raise NotImplementedError


class UnmarshalerFactoryMethodClass(UnmarshalerFactory, lang.Abstract):
    @gfs.method(instance_cache=True)
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        raise NotImplementedError



import abc
import functools
import operator
import typing as ta

from .. import check
from .. import collections as col
from .. import lang


T = ta.TypeVar('T')
T_co = ta.TypeVar('T_co', covariant=True)
U = ta.TypeVar('U')
P = ta.ParamSpec('P')


##


class GuardFn(ta.Protocol[P, T_co]):
    def __get__(self, instance, owner=None): ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T_co] | None: ...


##


@ta.final
class DumbGuardFn(ta.Generic[P, T]):
    def __init__(self, fn: ta.Callable[P, T]) -> None:
        self._fn = fn

    def __get__(self, instance, owner=None):
        return DumbGuardFn(self._fn.__get__(instance, owner))  # noqa

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T]:
        return functools.partial(self._fn, *args, **kwargs)


dumb = DumbGuardFn


##


class AmbiguousGuardFnError(Exception):
    pass


@ta.final
class MultiGuardFn(ta.Generic[P, T]):
    def __init__(
            self,
            *children: GuardFn[P, T],
            default: GuardFn[P, T] | None = None,
            strict: bool = False,
    ) -> None:
        self._children, self._default, self._strict = children, default, strict

    lang.attr_ops(lambda self: (
        self._children,
        self._default,
        self._strict,
    )).install(locals())

    def __get__(self, instance, owner=None):
        return MultiGuardFn(*map(operator.methodcaller('__get__', instance, owner), self._children), default=self._default.__get__(instance, owner) if self._default is not None else None, strict=self._strict)  # noqa

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T] | None:
        matches = []
        for c in self._children:
            if (m := c(*args, **kwargs)) is not None:
                if not self._strict:
                    return m
                matches.append(m)
        if not matches:
            if (dfl := self._default) is not None:
                return dfl(*args, **kwargs)
            else:
                return None
        elif len(matches) > 1:
            raise AmbiguousGuardFnError
        else:
            return matches[0]


multi = MultiGuardFn


##


class _BaseGuardFnMethod(lang.Abstract, ta.Generic[P, T]):
    def __init__(
            self,
            *,
            strict: bool = False,
            requires_override: bool = False,
            instance_cache: bool = False,
            default: GuardFn[P, T] | None = None,
    ) -> None:
        super().__init__()

        self._strict = strict
        self._instance_cache = instance_cache
        self._default = default

        self._registry: col.AttrRegistry[ta.Callable, None] = col.AttrRegistry(
            requires_override=requires_override,
        )

        self._cache: col.AttrRegistryCache[ta.Callable, None, MultiGuardFn] = col.AttrRegistryCache(
            self._registry,
            self._prepare,
        )

    _owner: type | None = None
    _name: str | None = None

    def __set_name__(self, owner, name):
        if self._owner is None:
            self._owner = owner
        if self._name is None:
            self._name = name

    def register(self, fn: U) -> U:
        check.callable(fn)
        self._registry.register(ta.cast(ta.Callable, fn), None)
        return fn

    def _prepare(self, instance_cls: type, collected: ta.Mapping[str, tuple[ta.Callable, None]]) -> MultiGuardFn:
        return MultiGuardFn(
            *[getattr(instance_cls, a) for a in collected],
            default=self._default,
            strict=self._strict,
        )

    @abc.abstractmethod
    def _bind(self, instance, owner):
        raise NotImplementedError

    def __get__(self, instance, owner=None):
        if instance is None:
            return self

        if self._instance_cache:
            try:
                return instance.__dict__[self._name]
            except KeyError:
                pass

        bound = self._bind(instance, owner)

        if self._instance_cache:
            instance.__dict__[self._name] = bound

        return bound

    def _call(self, *args, **kwargs):
        instance, *rest = args
        return self.__get__(instance)(*rest, **kwargs)

#


@ta.final
class GuardFnMethod(_BaseGuardFnMethod[P, T]):
    def _bind(self, instance, owner):
        return self._cache.get(type(instance)).__get__(instance, owner)  # noqa

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T] | None:
        return self._call(*args, **kwargs)


def method(
        *,
        strict: bool = False,
        requires_override: bool = False,
        instance_cache: bool = False,
        default: bool = False,
) -> ta.Callable[[ta.Callable[P, T]], GuardFnMethod[P, T]]:  # noqa
    def inner(fn):
        return GuardFnMethod(
            strict=strict,
            requires_override=requires_override,
            instance_cache=instance_cache,
            default=fn if default else None,
        )

    return inner


#


@ta.final
class ImmediateGuardFnMethod(_BaseGuardFnMethod[P, T]):
    def _bind(self, instance, owner):
        gf = self._cache.get(type(instance)).__get__(instance, owner)  # noqa

        def inner(*args, **kwargs):
            return gf(*args, **kwargs)()  # Note: cannot be None due to non-optional default

        return inner

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self._call(*args, **kwargs)


def immediate_method(
        *,
        strict: bool = False,
        requires_override: bool = False,
        instance_cache: bool = False,
) -> ta.Callable[[ta.Callable[P, T]], ImmediateGuardFnMethod[P, T]]:  # noqa
    def inner(fn):
        return ImmediateGuardFnMethod(
            strict=strict,
            requires_override=requires_override,
            instance_cache=instance_cache,
            default=(lambda *args, **kwargs: lambda: fn(*args, **kwargs)),
        )

    return inner



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import asyncio
import typing as ta

from ...lite.timeouts import Timeout
from ...lite.timeouts import TimeoutLike


AwaitableT = ta.TypeVar('AwaitableT', bound=ta.Awaitable)


##


def asyncio_maybe_timeout(
        fut: AwaitableT,
        timeout: TimeoutLike = None,
) -> AwaitableT:
    if timeout is not None:
        fut = asyncio.wait_for(fut, Timeout.of(timeout)())  # type: ignore
    return fut



# ruff: noqa: UP006 UP043 UP045
# @omlish-lite
"""
TODO:
 - sync (lol) w/ asyncs.anyio
 - atomics
 - Once poison=False, PoisonedError
"""
import collections
import threading
import typing as ta

from .lite.maybes import Maybe


T = ta.TypeVar('T')


##


class SyncOnce:
    def __init__(self) -> None:
        super().__init__()

        self._done = False
        self._lock = threading.Lock()

    def do(self, fn: ta.Callable[[], None]) -> bool:
        if self._done:
            return False
        with self._lock:
            if self._done:
                return False  # type: ignore
            try:
                fn()
            finally:
                self._done = True
            return True


##


class SyncLazy(ta.Generic[T]):
    def __init__(self) -> None:
        super().__init__()

        self._once = SyncOnce()
        self._v: Maybe[T] = Maybe.empty()

    def peek(self) -> Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = Maybe.just(v)

    def get(self, fn: ta.Callable[[], T]) -> T:
        def do():
            self._v = Maybe.just(fn())
        self._once.do(do)
        return self._v.must()


class SyncLazyFn(ta.Generic[T]):
    def __init__(self, fn: ta.Callable[[], T]) -> None:
        super().__init__()

        self._fn = fn
        self._once = SyncOnce()
        self._v: Maybe[T] = Maybe.empty()

    def peek(self) -> Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = Maybe.just(v)

    def get(self) -> T:
        def do():
            self._v = Maybe.just(self._fn())
        self._once.do(do)
        return self._v.must()


##


class ConditionDeque(ta.Generic[T]):
    def __init__(
            self,
            *,
            cond: ta.Optional['threading.Condition'] = None,
            deque: ta.Optional[ta.Deque[T]] = None,

            lock: ta.Optional['threading.RLock'] = None,
            maxlen: ta.Optional[int] = None,
            init: ta.Optional[ta.Iterable[T]] = None,
    ) -> None:
        super().__init__()

        if cond is None:
            cond = threading.Condition(lock=lock)
        if deque is None:
            deque = collections.deque(maxlen=maxlen)
        if init:
            deque.extend(init)

        self._cond = cond
        self._deque = deque

    @property
    def cond(self) -> 'threading.Condition':
        return self._cond

    @property
    def deque(self) -> ta.Deque[T]:
        return self._deque

    def push(
            self,
            *items: T,
            n: int = 1,
    ) -> None:
        with self.cond:
            self.deque.extend(items)
            self.cond.notify(n)

    def pop(
            self,
            timeout: ta.Optional[float] = None,
            *,
            if_empty: ta.Optional[ta.Callable[[], None]] = None,
    ) -> T:
        with self.cond:
            if not self.deque and if_empty is not None:
                if_empty()
            while not self.deque:
                self.cond.wait(timeout)
            return self.deque.popleft()


##


class SyncBufferRelay(ta.Generic[T]):
    def __init__(
            self,
            *,
            wake_fn: ta.Callable[[], None],
    ) -> None:
        super().__init__()

        self._wake_fn = wake_fn

        self._lock = threading.Lock()
        self._buffer: SyncBufferRelay._Buffer = SyncBufferRelay._Buffer()

    class _Buffer:
        def __init__(self) -> None:
            self.lst: list = []
            self.age = 0

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}({self.lst!r}, age={self.age!r})'

    def push(self, *vs: T) -> None:
        with self._lock:
            buf = self._buffer
            needs_wake = not buf.age
            buf.lst.extend(vs)
            buf.age += 1
        if needs_wake:
            self._wake_fn()

    def swap(self) -> ta.Sequence[T]:
        with self._lock:
            buf, self._buffer = self._buffer, SyncBufferRelay._Buffer()
        return buf.lst


##


class CountDownLatch:
    def __init__(self, count: int) -> None:
        super().__init__()

        self._count = count
        self._cond = threading.Condition()

    def count_down(self) -> None:
        with self._cond:
            self._count -= 1
            if self._count <= 0:
                self._cond.notify_all()

    def wait(
            self,
            timeout: ta.Optional[float] = None,
    ) -> bool:
        with self._cond:
            return self._cond.wait_for(
                lambda: self._count < 1,
                timeout,
            )



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import typing as ta

from ..lite.abstract import Abstract
from .addresses import SocketAddress
from .io import SocketIoPair


SocketHandler = ta.Callable[[SocketAddress, 'SocketIoPair'], None]  # ta.TypeAlias


##


class SocketHandler_(Abstract):  # noqa
    @abc.abstractmethod
    def __call__(self, addr: SocketAddress, f: SocketIoPair) -> None:
        raise NotImplementedError


class SocketHandlerClose(Exception):  # noqa
    pass



# ruff: noqa: UP006
import functools
import typing as ta


##


_ANNOTATION_ATTRS: ta.FrozenSet[str] = frozenset([
    '__annotations__',

    '__annotate__',
    '__annotate_func__',

    '__annotations_cache__',
])

_UPDATE_WRAPPER_ASSIGNMENTS_NO_ANNOTATIONS: ta.Sequence[str] = list(frozenset(functools.WRAPPER_ASSIGNMENTS) - _ANNOTATION_ATTRS)  # noqa


def update_wrapper_no_annotations(wrapper, wrapped):
    functools.update_wrapper(wrapper, wrapped, assigned=_UPDATE_WRAPPER_ASSIGNMENTS_NO_ANNOTATIONS)
    return wrapper



"""
Insufficient alt impls:
 - `sys.meta_path`:
  - need access to the `fromlist` and `level` arguments passed to `__import__`
  - need to return fake modules to the import operation which are not added to `sys.modules`
 - `sys.addaudithook`: cannot prevent import or inject result
 - `sys.settrace` / bytecode tracing: same
 - jit bytecode rewriting: slower than just importing everything

Possible alt impls:
 - aot static analysis, codegen, compare, if valid skip ctxmgr body and inject proxies, otherwise warn and run
  - or just gen code inline, if ta.TYPE_CHECKING: .... else: # @omlish-generate-auto-proxy-import/init
   - generate classic `foo = _lang.proxy_import('.foo', __package__)` blocks
"""
import builtins
import contextlib
import functools
import importlib.util
import sys
import threading
import types
import typing as ta


##


class ImportCaptureError(Exception):
    pass


class ImportCaptureErrors:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    class HookError(ImportCaptureError):
        pass

    class AttrError(ImportCaptureError):
        def __init__(self, module: str | None, name: str) -> None:
            super().__init__()

            self.module = module
            self.name = name

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(module={self.module!r}, name={self.name!r})'

    class ImportError(ImportCaptureError):  # noqa
        def __init__(
                self,
                name: str,
                *,
                level: int | None = None,
                from_list: ta.Sequence[str] | None,
        ) -> None:
            super().__init__()

            self.name = name
            self.level = level
            self.from_list = from_list

        def __repr__(self) -> str:
            return ''.join([
                f'{self.__class__.__qualname__}(',
                f'name={self.name!r}',
                *([f', level={self.level!r}'] if self.level is not None else []),
                *([f', from_list={self.from_list!r}'] if self.from_list is not None else []),
                ')',
            ])

    class ImportStarForbiddenError(ImportError):
        pass

    class UncapturedImportForbiddenError(ImportError):
        pass

    class UnreferencedImportsError(ImportCaptureError):
        def __init__(self, unreferenced: ta.Sequence[str]) -> None:
            super().__init__()

            self.unreferenced = unreferenced

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}(unreferenced={self.unreferenced!r})'

    class CaptureInProgressError(ImportCaptureError):
        pass


##


class _ImportCaptureHook:
    def __init__(
            self,
            *,
            package: str | None = None,
            forbid_uncaptured_imports: bool = False,
    ) -> None:
        super().__init__()

        self._package = package
        self._forbid_uncaptured_imports = forbid_uncaptured_imports

        self._modules_by_name: dict[str, _ImportCaptureHook._Module] = {}
        self._modules_by_module_obj: dict[types.ModuleType, _ImportCaptureHook._Module] = {}

    #

    class _Module:
        def __init__(
                self,
                name: str,
                getattr_handler: ta.Callable[['_ImportCaptureHook._Module', str], ta.Any],
                *,
                parent: ta.Optional['_ImportCaptureHook._Module'] = None,
        ) -> None:
            super().__init__()

            if name.startswith('.'):
                raise ImportCaptureError

            self.name = name
            self.parent = parent

            self.base_name = name.rpartition('.')[2]
            self.root: _ImportCaptureHook._Module = parent.root if parent is not None else self  # noqa

            self.children: dict[str, _ImportCaptureHook._Module] = {}
            self.descendants: set[_ImportCaptureHook._Module] = set()

            self.module_obj = types.ModuleType(f'<{self.__class__.__qualname__}: {name}>')
            self.module_obj.__file__ = None
            self.module_obj.__getattr__ = functools.partial(getattr_handler, self)  # type: ignore[method-assign]  # noqa
            self.initial_module_dict = dict(self.module_obj.__dict__)

            self.explicit = False
            self.immediate = False

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}<{self.name}{"!" if self.immediate else "+" if self.explicit else ""}>'

        def set_explicit(self) -> None:
            cur: _ImportCaptureHook._Module | None = self
            while cur is not None and not cur.explicit:
                cur.explicit = True
                cur = cur.parent

    #

    @property
    def _modules(self) -> ta.Sequence[_Module]:
        return sorted(self._modules_by_name.values(), key=lambda m: m.name)

    def _get_or_make_module(self, name: str) -> _Module:
        try:
            return self._modules_by_name[name]
        except KeyError:
            pass

        parent: _ImportCaptureHook._Module | None = None
        if '.' in name:
            rest, _, attr = name.rpartition('.')
            parent = self._get_or_make_module(rest)
            if attr in parent.children:
                raise ImportCaptureErrors.AttrError(rest, attr)

        module = _ImportCaptureHook._Module(
            name,
            self._handle_module_getattr,
            parent=parent,
        )
        self._modules_by_name[name] = module
        self._modules_by_module_obj[module.module_obj] = module

        if parent is not None:
            parent.children[module.base_name] = module
            setattr(parent.module_obj, module.base_name, module.module_obj)
            parent.root.descendants.add(module)

        return module

    def _make_child_module(self, module: _Module, attr: str) -> _Module:
        if attr in module.children:
            raise ImportCaptureErrors.AttrError(module.name, attr)

        return self._get_or_make_module(f'{module.name}.{attr}')

    #

    def _handle_module_getattr(self, module: _Module, attr: str) -> ta.Any:
        if not module.explicit:
            raise ImportCaptureErrors.AttrError(module.name, attr)

        return self._make_child_module(module, attr).module_obj

    def _handle_import(
            self,
            name: str,
            *,
            from_list: ta.Sequence[str] | None,
    ) -> types.ModuleType:
        module = self._get_or_make_module(name)

        if from_list is None:
            module.set_explicit()
            module.root.immediate = True
            return module.root.module_obj

        else:
            for attr in from_list:
                if attr == '*':
                    raise ImportCaptureErrors.ImportStarForbiddenError(module.name, from_list=from_list)

                if (cm := module.children.get(attr)) is None:
                    cm = self._make_child_module(module, attr)
                    cm.set_explicit()
                    cm.immediate = True
                    continue

                x = getattr(module.module_obj, attr)
                if x is not cm.module_obj or x not in self._modules_by_module_obj:
                    raise ImportCaptureErrors.AttrError(module.name, attr)

            return module.module_obj

    #

    _MOD_SELF_ATTR: ta.ClassVar[str] = '__import_capture__'

    def _intercept_import(
            self,
            name: str,
            *,
            globals: ta.Mapping[str, ta.Any] | None = None,  # noqa
            from_list: ta.Sequence[str] | None = None,
            level: int = 0,
    ) -> types.ModuleType | None:
        if not (
                globals is not None and
                globals.get(self._MOD_SELF_ATTR) is self
        ):
            return None

        if level:
            if not self._package:
                raise ImportCaptureError
            name = importlib.util.resolve_name(('.' * level) + name, self._package)

        return self._handle_import(
            name,
            from_list=from_list,
        )

    @ta.final
    @contextlib.contextmanager
    def hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.Iterator[None]:
        if self._MOD_SELF_ATTR in mod_globals:
            raise ImportCaptureErrors.HookError

        mod_globals[self._MOD_SELF_ATTR] = self

        try:
            with self._hook_context(mod_globals):
                yield

        finally:
            if mod_globals[self._MOD_SELF_ATTR] is not self:
                raise ImportCaptureErrors.HookError

            del mod_globals[self._MOD_SELF_ATTR]

    # @abc.abstractmethod
    def _hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.ContextManager[None]:
        raise NotImplementedError

    #

    def verify_state(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> None:
        for m in self._modules_by_name.values():
            if m.immediate and not m.explicit:
                raise ImportCaptureError

            if not m.explicit and m.children:
                raise ImportCaptureError

            for a, o in m.module_obj.__dict__.items():
                try:
                    i = m.initial_module_dict[a]

                except KeyError:
                    if o is not m.children[a].module_obj:
                        raise ImportCaptureErrors.AttrError(m.name, a) from None

                else:
                    if o != i:
                        raise ImportCaptureErrors.AttrError(m.name, a)

    #

    def build_captured(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
            *,
            collect_unreferenced: bool = False,
    ) -> 'ImportCapture.Captured':
        rem_explicit_mods: set[_ImportCaptureHook._Module] = set()
        if collect_unreferenced:
            rem_explicit_mods.update(
                m for m in self._modules_by_name.values()
                if m.immediate
                and m.parent is not None  # No good way to tell if user did `import a.b.c` or `import a.b.c as c`
            )

        #

        dct: dict[_ImportCaptureHook._Module, list[tuple[str | None, str]]] = {}

        for attr, obj in mod_globals.items():
            if isinstance(obj, _ImportCaptureHook._Module):
                raise ImportCaptureErrors.AttrError(None, attr) from None

            elif isinstance(obj, types.ModuleType):
                try:
                    m = self._modules_by_module_obj[obj]
                except KeyError:
                    continue

                if m.explicit:
                    dct.setdefault(m, []).append((None, attr))
                    if m in rem_explicit_mods:
                        # Remove everything reachable from this root *except* items imported immediately, such as
                        # `from x import y` - those still need to be immediately reachable.
                        rem_explicit_mods -= {dm for dm in m.descendants if not dm.immediate}
                        rem_explicit_mods.remove(m)

                else:
                    p = m.parent
                    if p is None or not p.explicit:
                        raise ImportCaptureError
                    dct.setdefault(p, []).append((m.base_name, attr))

        #

        mods: dict[str, ImportCapture.Module] = {}

        def build_import_module(m: _ImportCaptureHook._Module) -> ImportCapture.Module:
            children: dict[str, ImportCapture.Module] = {}
            attrs: list[str] = []
            for cm in sorted(m.children.values(), key=lambda cm: cm.name):
                if not cm.explicit:
                    attrs.append(cm.base_name)
                else:
                    children[cm.base_name] = build_import_module(cm)

            mod = ImportCapture.Module(
                m.name,
                children or None,
                attrs or None,
            )

            if m.parent is None:
                mod.parent = None
            for c in children.values():
                c.parent = mod

            mods[mod.name] = mod
            return mod

        root_mods: dict[str, ImportCapture.Module] = {
            m.base_name: build_import_module(m)
            for m in self._modules_by_name.values()
            if m.parent is None
        }

        mods = dict(sorted(mods.items(), key=lambda t: t[0]))
        root_mods = dict(sorted(root_mods.items(), key=lambda t: t[0]))

        #

        imps: list[ImportCapture.Import] = []

        for m, ts in sorted(dct.items(), key=lambda t: t[0].name):
            imps.append(ImportCapture.Import(
                mods[m.name],
                [r for l, r in ts if l is None] or None,
                [(l, r) for l, r in ts if l is not None] or None,
            ))

        #

        unreferenced: list[str] | None = None
        if collect_unreferenced and rem_explicit_mods:
            unreferenced = sorted(m.name for m in rem_explicit_mods)

        return ImportCapture.Captured(
            {i.module.name: i for i in imps},

            mods,
            root_mods,

            unreferenced,
        )


#


class _AbstractBuiltinsImportCaptureHook(_ImportCaptureHook):
    def __init__(
            self,
            *,
            _frame: types.FrameType | None = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

    def _new_import(
            self,
            old_import,
            name,
            globals=None,  # noqa
            locals=None,  # noqa
            fromlist=None,
            level=0,
    ):
        if (im := self._intercept_import(
                name,
                globals=globals,
                from_list=fromlist,
                level=level,
        )) is not None:
            return im

        if self._forbid_uncaptured_imports:
            raise ImportCaptureErrors.UncapturedImportForbiddenError(
                name,
                level=level,
                from_list=fromlist,
            )

        return old_import(
            name,
            globals=globals,
            locals=locals,
            fromlist=fromlist,
            level=level,
        )


#


class _UnsafeGlobalBuiltinsImportCaptureHook(_AbstractBuiltinsImportCaptureHook):
    @contextlib.contextmanager
    def _hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.Iterator[None]:
        old_import = builtins.__import__
        new_import = functools.partial(self._new_import, old_import)

        builtins.__import__ = new_import

        try:
            yield

        finally:
            if builtins.__import__ is not new_import:
                raise ImportCaptureErrors.HookError

            builtins.__import__ = old_import


class _SomewhatThreadSafeGlobalBuiltinsImportCaptureHook(_AbstractBuiltinsImportCaptureHook):
    class _AlreadyPatchedError(Exception):
        pass

    @ta.final
    class _Patch:
        __lock: ta.ClassVar[threading.Lock] = threading.Lock()

        def __init__(self, old_import):
            self.__old_import = old_import
            self.__hooks = {}
            self.__uninstalled = False

        @classmethod
        def _add_hook(cls, mod_globals, new_import) -> '_SomewhatThreadSafeGlobalBuiltinsImportCaptureHook._Patch':
            gi = id(mod_globals)
            for _ in range(1_000):
                try:
                    with cls.__lock:
                        x: ta.Any = builtins.__import__
                        p: _SomewhatThreadSafeGlobalBuiltinsImportCaptureHook._Patch
                        if x.__class__ is cls:
                            p = x
                            if p.__uninstalled:  # noqa
                                raise _SomewhatThreadSafeGlobalBuiltinsImportCaptureHook._AlreadyPatchedError  # noqa
                        else:
                            p = cls(x)
                            builtins.__import__ = p
                        p.__hooks[gi] = (mod_globals, new_import)
                        return p
                except _SomewhatThreadSafeGlobalBuiltinsImportCaptureHook._AlreadyPatchedError:
                    pass
            raise ImportCaptureErrors.HookError('Failed to install builtins hook')

        def _remove_hook(self, mod_globals, *, no_raise=False):
            gi = id(mod_globals)
            with self.__lock:
                tg, _ = self.__hooks[gi]
                del self.__hooks[gi]
                if not self.__uninstalled and not self.__hooks:
                    self.__uninstalled = True
                    if builtins.__import__ is not self:
                        if not no_raise:
                            # TODO: warn?
                            raise ImportCaptureErrors.HookError('Unexpected builtins hook')
                    else:
                        builtins.__import__ = self.__old_import
                if tg is not mod_globals:
                    if not no_raise:
                        # TODO: warn?
                        raise ImportCaptureErrors.HookError('Mismatched globals')

        def __call__(
                self,
                name,
                globals=None,  # noqa
                locals=None,  # noqa
                fromlist=None,
                level=0,
        ):
            if globals is not None and (tup := self.__hooks.get(id(globals))) is not None:
                tg, tf = tup
                if tg is globals:
                    return tf(
                        self.__old_import,
                        name,
                        globals=globals,
                        locals=locals,
                        fromlist=fromlist,
                        level=level,
                    )
                else:
                    self._remove_hook(tg, no_raise=True)

            return self.__old_import(
                name,
                globals=globals,
                locals=locals,
                fromlist=fromlist,
                level=level,
            )

    @contextlib.contextmanager
    def _hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.Iterator[None]:
        patch = _SomewhatThreadSafeGlobalBuiltinsImportCaptureHook._Patch._add_hook(mod_globals, self._new_import)  # noqa

        try:
            yield

        finally:
            patch._remove_hook(mod_globals)  # noqa


#


_cext_: ta.Any


def _cext() -> ta.Any:
    global _cext_
    try:
        return _cext_
    except NameError:
        pass

    cext: ta.Any
    try:
        from . import _capture as cext  # type: ignore
    except ImportError:
        cext = None

    _cext_ = cext
    return cext


class _FrameBuiltinsImportCaptureHook(_AbstractBuiltinsImportCaptureHook):
    def __init__(
            self,
            *,
            _frame: types.FrameType,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(**kwargs)

        self._frame = _frame

    @classmethod
    def _set_frame_builtins(
            cls,
            frame: types.FrameType,
            new_builtins: dict[str, ta.Any],
    ) -> bool:
        return _cext()._set_frame_builtins(frame, frame.f_builtins, new_builtins)  # noqa

    @contextlib.contextmanager
    def _hook_context(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
    ) -> ta.Iterator[None]:
        old_builtins = self._frame.f_builtins
        old_import = old_builtins['__import__']
        new_import = functools.partial(self._new_import, old_import)

        new_builtins = dict(old_builtins)
        new_builtins['__import__'] = new_import
        if not self._set_frame_builtins(self._frame, new_builtins):
            raise ImportCaptureErrors.HookError

        try:
            yield

        finally:
            if self._frame.f_builtins is not new_builtins:
                raise ImportCaptureErrors.HookError

            if not self._set_frame_builtins(self._frame, old_builtins):
                raise ImportCaptureErrors.HookError


#


_CAPTURE_IMPLS: ta.Mapping[str, type[_AbstractBuiltinsImportCaptureHook]] = {
    'cext': _FrameBuiltinsImportCaptureHook,
    'somewhat_safe': _SomewhatThreadSafeGlobalBuiltinsImportCaptureHook,
    'unsafe': _UnsafeGlobalBuiltinsImportCaptureHook,
}


def _new_import_capture_hook(
        mod_globals: ta.MutableMapping[str, ta.Any],  # noqa
        *,
        stack_offset: int = 0,
        capture_impl: str | None = None,
        **kwargs: ta.Any,
) -> '_ImportCaptureHook':
    if '_frame' not in kwargs:
        frame: types.FrameType | None = sys._getframe(1 + stack_offset)  # noqa
        if frame is None or frame.f_globals is not mod_globals:
            raise ImportCaptureError("Can't find importing frame")
        kwargs['_frame'] = frame

    kwargs.setdefault('package', mod_globals.get('__package__'))

    cls: type[_AbstractBuiltinsImportCaptureHook]
    if capture_impl is not None:
        cls = _CAPTURE_IMPLS[capture_impl]
    elif _cext() is not None:
        cls = _FrameBuiltinsImportCaptureHook
    else:
        cls = _SomewhatThreadSafeGlobalBuiltinsImportCaptureHook

    return cls(**kwargs)


##


ImportCaptureModuleKind: ta.TypeAlias = ta.Literal[
    'parent',
    'terminal',
    'leaf',
]


class ImportCapture:
    @ta.final
    class Module:
        def __init__(
                self,
                name: str,
                children: ta.Mapping[str, 'ImportCapture.Module'] | None = None,
                attrs: ta.Sequence[str] | None = None,
        ) -> None:
            self.name = name
            self.children = children
            self.attrs = attrs

            self.base_name = name.rpartition('.')[2]

            if not self.children and not self.attrs:
                self.kind = 'leaf'
            elif not self.children or all(c.kind == 'leaf' for c in self.children.values()):
                self.kind = 'terminal'
            else:
                self.kind = 'parent'

        parent: ta.Optional['ImportCapture.Module']

        kind: ImportCaptureModuleKind

        def __repr__(self) -> str:
            return ''.join([
                f'{self.__class__.__name__}(',
                f'{self.name!r}',
                f', :{self.kind}',
                *([f', children=[{", ".join(map(repr, self.children))}]'] if self.children else []),
                *([f', attrs={self.attrs!r}'] if self.attrs else []),
                ')',
            ])

        _root: 'ImportCapture.Module'

        @property
        def root(self) -> 'ImportCapture.Module':
            try:
                return self._root
            except AttributeError:
                pass

            root = self
            while root.parent is not None:
                root = root.parent
            self._root = root
            return root

    @ta.final
    class Import:
        def __init__(
                self,
                module: 'ImportCapture.Module',
                as_: ta.Sequence[str] | None,
                attrs: ta.Sequence[tuple[str, str]] | None,  # ('foo', 'bar') -> `import foo as bar` - explicitly not a dict  # noqa
        ) -> None:
            self.module = module
            self.as_ = as_
            self.attrs = attrs

        def __repr__(self) -> str:
            return ''.join([
                f'{self.__class__.__name__}(',
                f'{self.module.name!r}',
                *([f', as_={self.as_!r}'] if self.as_ else []),
                *([f', attrs={self.attrs!r}'] if self.attrs else []),
                ')',
            ])

    @ta.final
    class Captured:
        def __init__(
                self,

                imports: ta.Mapping[str, 'ImportCapture.Import'],

                modules: ta.Mapping[str, 'ImportCapture.Module'],
                root_modules: ta.Mapping[str, 'ImportCapture.Module'],

                unreferenced: ta.Sequence[str] | None,
        ) -> None:
            self.imports = imports

            self.modules = modules
            self.root_modules = root_modules

            self.unreferenced = unreferenced

        @property
        def attrs(self) -> ta.Iterator[str]:
            for pi in self.imports.values():
                if pi.as_:
                    yield from pi.as_
                if pi.attrs:
                    for _, a in pi.attrs:
                        yield a

    EMPTY_CAPTURED: ta.ClassVar[Captured] = Captured(
        {},
        {},
        {},
        None,
    )

    #

    def __init__(
            self,
            mod_globals: ta.MutableMapping[str, ta.Any],
            *,
            _hook: _ImportCaptureHook,

            disable: bool = False,
    ) -> None:
        super().__init__()

        self._mod_globals = mod_globals
        self._hook = _hook

        self._disabled = disable

    @property
    def disabled(self) -> bool:
        return self._disabled

    #

    class _Result(ta.NamedTuple):
        captured: 'ImportCapture.Captured'

    _result_: _Result | None = None

    @property
    def _result(self) -> _Result:
        if (rs := self._result_) is None:
            raise ImportCaptureErrors.CaptureInProgressError
        return rs

    @property
    def is_complete(self) -> bool:
        return self._result_ is not None

    @property
    def captured(self) -> Captured:
        return self._result.captured

    #

    @contextlib.contextmanager
    def capture(
            self,
            *,
            unreferenced_callback: ta.Callable[[ta.Sequence[str]], None] | None = None,
            raise_unreferenced: bool = False,
    ) -> ta.Iterator[ta.Self]:
        if self._result_ is not None:
            raise ImportCaptureError('capture already complete')

        if self._disabled:
            self._result_ = ImportCapture._Result(
                ImportCapture.EMPTY_CAPTURED,
            )
            yield self
            return

        with self._hook.hook_context(self._mod_globals):
            yield self

        self._hook.verify_state(self._mod_globals)

        blt = self._hook.build_captured(
            self._mod_globals,
            collect_unreferenced=unreferenced_callback is not None or raise_unreferenced,
        )

        if blt.unreferenced:
            if unreferenced_callback:
                unreferenced_callback(blt.unreferenced)
            if raise_unreferenced:
                raise ImportCaptureErrors.UnreferencedImportsError(blt.unreferenced)

        for a in blt.attrs:
            del self._mod_globals[a]

        self._result_ = ImportCapture._Result(
            blt,
        )

    #

    def update_exports(self) -> None:
        cap = self._result.captured

        try:
            al: ta.Any = self._mod_globals['__all__']
        except KeyError:
            al = self._mod_globals['__all__'] = [k for k in self._mod_globals if not k.startswith('_')]
        else:
            if not isinstance(al, ta.MutableSequence):
                al = self._mod_globals['__all__'] = list(al)

        al_s = set(al)
        for a in cap.attrs:
            if a not in al_s:
                al.append(a)
                al_s.add(a)



import typing as ta

from .... import lang
from ..processing.base import ProcessingContext
from ..processing.registry import register_processing_context_item_factory


##


MroDict = ta.NewType('MroDict', ta.Mapping[str, ta.Any])


@register_processing_context_item_factory(MroDict)
def _mro_dict_processing_context_item_factory(ctx: ProcessingContext) -> MroDict:
    return MroDict(lang.mro_dict(ctx.cls))
